
/*******************************************************************************
*
* File check4.c
*
* Copyright (C) 2005 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Check of the block grid allocation program
*
*******************************************************************************/

#define MAIN_PROGRAM

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "mpi.h"
#include "su3.h"
#include "flags.h"
#include "random.h"
#include "start.h"
#include "block.h"
#include "global.h"

static int itest[VOLUME];


static void set_itest2zero(void)
{
   int ix;

   for (ix=0;ix<VOLUME;ix++)
      itest[ix]=0;
}


static void add2itest(block_t *b)
{
   int ix,iy;

   for (ix=0;ix<(*b).vol;ix++)
   {
      iy=(*b).imb[ix];
      itest[iy]+=1;
   }
}


static int check_itest(void)
{
   int ix;

   for (ix=0;ix<VOLUME;ix++)
   {
      if (itest[ix]!=1)
         return 1;
   }

   return 0;
}


int main(int argc,char *argv[])
{
   int my_rank,ifc,n,n0,n1,n2,n3;
   int bs[4],nb,isw;
   int iu,iud,ns,nsd,nsb,nsbd;
   int shu,shud,shs,shsd,shsb,shsbd;
   block_t *b,*b0;
   bndry_t *bb,*bb0;
   blk_grid_t grid;
   FILE *flog=NULL,*fin=NULL;

   MPI_Init(&argc,&argv);
   MPI_Comm_rank(MPI_COMM_WORLD,&my_rank);

   if (my_rank==0)
   {
      flog=freopen("check4.log","w",stdout);
      fin=freopen("check4.in","r",stdin);

      printf("\n");
      printf("Check of the block grid allocation program\n");
      printf("------------------------------------------\n\n");

      printf("%dx%dx%dx%d lattice, ",NPROC0*L0,NPROC1*L1,NPROC2*L2,NPROC3*L3);
      printf("%dx%dx%dx%d process grid, ",NPROC0,NPROC1,NPROC2,NPROC3);
      printf("%dx%dx%dx%d local lattice\n\n",L0,L1,L2,L3);

      scanf("bs %d %d %d %d",&bs[0],&bs[1],&bs[2],&bs[3]);
      fclose(fin);

      printf("bs = %d %d %d %d\n\n",bs[0],bs[1],bs[2],bs[3]);
   }

   MPI_Bcast(bs,4,MPI_INT,0,MPI_COMM_WORLD);

   start_ranlux(0,12345);
   geometry();
   alloc_ud();
   alloc_u();

   for (grid=0;grid<BLK_GRIDS;grid++)
   {
      alloc_bgr(grid,bs);

      print_blk_flags(grid,0);
      print_blk_flags(grid,1);

      b=blk_list(grid,&nb,&isw);

      n0=L0/bs[0];
      n1=L1/bs[1];
      n2=L2/bs[2];
      n3=L3/bs[3];
      n=n0*cpr[0]+n1*cpr[1]+n2*cpr[2]+n3*cpr[3];

      error((b==NULL)||(nb!=(n0*n1*n2*n3))||(isw!=(n%2)),1,
             "main [check4.c]","Incorrect return values of blk_list");

      if (grid==GCR_BLOCKS)
      {
         iu=1;
         iud=0;
         ns=3;
         nsd=0;
         nsb=0;
         nsbd=0;

         shu=0;
         shud=0;
         shs=1;
         shsd=0;
         shsb=0;
         shsbd=0;
      }
      else if (grid==HMC_BLOCKS)
      {
         iu=1;
         iud=1;
         ns=5;
         nsd=7;
         nsb=0;
         nsbd=1;

         shu=1;
         shud=0;
         shs=1;
         shsd=1;
         shsb=0;
         shsbd=1;
      }
      else if (grid==DFL_BLOCKS)
      {
         iu=0;
         iud=1;
         ns=1;
         nsd=1;
         nsb=0;
         nsbd=0;

         shu=0;
         shud=1;
         shs=0;
         shsd=0;
         shsb=0;
         shsbd=0;
      }
      else
      {
         iu=0;
         iud=0;
         ns=0;
         nsd=0;
         nsb=0;
         nsbd=0;

         shu=0;
         shud=0;
         shs=0;
         shsd=0;
         shsb=0;
         shsbd=0;
      }

      b0=b;
      set_itest2zero();

      for (n=0;n<nb;n++)
      {
         error(((*b).bs[0]!=bs[0])||((*b).bs[1]!=bs[1])||
               ((*b).bs[2]!=bs[2])||((*b).bs[3]!=bs[3]),1,
         "main [check4.c]","b.bs is incorrect");

         error(((*b).bo[0]<0)||(((*b).bo[0]+bs[0])>L0)||
               ((*b).bo[1]<0)||(((*b).bo[1]+bs[1])>L1)||
               ((*b).bo[2]<0)||(((*b).bo[2]+bs[2])>L2)||
               ((*b).bo[3]<0)||(((*b).bo[3]+bs[3])>L3),1,
         "main [check4.c]","b.bo is out of range");

         error((((*b).bo[0]%bs[0])!=0)||(((*b).bo[1]%bs[1])!=0)||
               (((*b).bo[2]%bs[2])!=0)||(((*b).bo[3]%bs[3])!=0),1,
         "main [check4.c]","b.bo is not a integer multiple of bs");

         n0=(*b).bo[0]/bs[0];
         n1=(*b).bo[1]/bs[1];
         n2=(*b).bo[2]/bs[2];
         n3=(*b).bo[3]/bs[3];

         isw=(n0+n1+n2+n3)%2;

         error(((isw==0)&&(n>=(nb/2)))||((isw==1)&&(n<(nb/2))),1,
                 "main [check4.c]",
                 "Blocks are not locally checker-board ordered");

         n0=((iu==0)&&((*b).u!=NULL))||((iu==1)&&((*b).u==NULL));
         n1=((iu==1)&&(shu==0)&&(n>0)&&((*b).u==(*b0).u));
         n2=((shu==1)&&((*b).u!=(*b0).u));

         error((n0==1)||(n1==1)||(n2==1),1,"main [check4.c]",
                "Incorrect value of b.u");

         n0=((iu==0)&&((*b).sw!=NULL))||((iu==1)&&((*b).sw==NULL));
         n1=((iu==1)&&(shu==0)&&(n>0)&&((*b).sw==(*b0).sw));
         n2=((shu==1)&&((*b).sw!=(*b0).sw));

         error((n0==1)||(n1==1)||(n2==1),1,"main [check4.c]",
                "Incorrect value of b.sw");

         n0=((iud==0)&&((*b).ud!=NULL))||((iud==1)&&((*b).ud==NULL));
         n1=((iud==1)&&(shud==0)&&(n>0)&&((*b).ud==(*b0).ud));
         n2=((shud==1)&&((*b).ud!=(*b0).ud));

         error((n0==1)||(n1==1)||(n2==1),1,"main [check4.c]",
                "Incorrect value of b.ud");

         n0=((iud==0)&&((*b).swd!=NULL))||((iud==1)&&((*b).swd==NULL));
         n1=((iud==1)&&(shud==0)&&(n>0)&&((*b).swd==(*b0).swd));
         n2=((shud==1)&&((*b).swd!=(*b0).swd));

         error((n0==1)||(n1==1)||(n2==1),1,"main [check4.c]",
                "Incorrect value of b.swd");

         n0=((ns==0)&&((*b).s!=NULL))||((ns>0)&&((*b).s==NULL));
         n1=((ns>0)&&(shs==0)&&(n>0)&&((*b).s==(*b0).s));
         n2=((shs==1)&&((*b).s!=(*b0).s));
         n3=(ns!=(*b).ns);

         error((n0==1)||(n1==1)||(n2==1)||(n3==1),1,"main [check4.c]",
                "Incorrect value of b.s");

         n0=((nsd==0)&&((*b).sd!=NULL))||((nsd>0)&&((*b).sd==NULL));
         n1=((nsd>0)&&(shsd==0)&&(n>0)&&((*b).sd==(*b0).sd));
         n2=((shsd==1)&&((*b).sd!=(*b0).sd));
         n3=(nsd!=(*b).nsd);

         error((n0==1)||(n1==1)||(n2==1)||(n3==1),1,"main [check4.c]",
                "Incorrect value of b.sd");

         add2itest(b);

         bb=(*b).bb;
         bb0=(*b0).bb;

         error ((*b).bb==NULL,1,
                  "main [check4.c]","Block boundaries are not allocated");

         for (ifc=0;ifc<8;ifc++)
         {
            n0=((iu==0)&&((*bb).u!=NULL))||((iu==1)&&((*bb).u==NULL));
            n1=((iu==1)&&(shu==0)&&(n>0)&&((*bb).u==(*bb0).u));
            n2=((shu==1)&&((*bb).u!=(*bb0).u));

            error((n0==1)||(n1==1)||(n2==1),1,"main [check4.c]",
                   "Incorrect value of bb.u");

            n0=((iud==0)&&((*bb).ud!=NULL))||((iud==1)&&((*bb).ud==NULL));
            n1=((iud==1)&&(shud==0)&&(n>0)&&((*bb).ud==(*bb0).ud));
            n2=((shud==1)&&((*bb).ud!=(*bb0).ud));

            error((n0==1)||(n1==1)||(n2==1),1,"main [check4.c]",
                   "Incorrect value of bb.ud");

            n0=((nsb==0)&&((*bb).s!=NULL))||((nsb>0)&&((*bb).s==NULL));
            n1=((nsb>0)&&(shsb==0)&&(n>0)&&((*bb).s==(*bb0).s));
            n2=((shsb==1)&&((*bb).s!=(*bb0).s));
            n3=(nsb!=(*bb).ns);

            error((n0==1)||(n1==1)||(n2==1)||(n3==1),1,"main [check4.c]",
                   "Incorrect value of bb.s");

            n0=((nsbd==0)&&((*bb).sd!=NULL))||((nsbd>0)&&((*bb).sd==NULL));
            n1=((nsbd>0)&&(shsbd==0)&&(n>0)&&((*bb).sd==(*bb0).sd));
            n2=((shsbd==1)&&((*bb).sd!=(*bb0).sd));
            n3=(nsbd!=(*bb).nsd);

            error((n0==1)||(n1==1)||(n2==1)||(n3==1),1,"main [check4.c]",
                   "Incorrect value of bb.sd");

            bb+=1;
            bb0+=1;
         }

         b+=1;
      }

      error(check_itest()!=0,1,"main [check4.c]",
            "Blocks do not cover the lattice with no overlaps");
   }

   error_chk();

   if (my_rank==0)
   {
      printf("No errors detected\n\n");
      fclose(flog);
   }

   MPI_Finalize();
   exit(0);
}
