
/*******************************************************************************
*
* File check1.c
*
* Copyright (C) 2007 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Check of the index arrays associated to the DFL_BLOCKS grid 
*
*******************************************************************************/

#define MAIN_PROGRAM

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "mpi.h"
#include "su3.h"
#include "flags.h"
#include "start.h"
#include "misc.h"
#include "dfl.h"
#include "global.h"

static int nb,(*inn)[8],*ipp;
static int nbb,nbbe[8],nbbo[8],ofsbbe[8],ofsbbo[8];


int main(int argc,char *argv[])
{
   int my_rank,bs[4],nbs,isw;
   int ix,iy,iz,ifc,ie,*bo1,*bo2;
   int l[4],mu,is;
   block_t *b;
   FILE *flog=NULL,*fin=NULL;

   MPI_Init(&argc,&argv);
   MPI_Comm_rank(MPI_COMM_WORLD,&my_rank);

   if (my_rank==0)
   {
      flog=freopen("check1.log","w",stdout);
      fin=freopen("check1.in","r",stdin);      
      
      printf("\n");
      printf("Check of the index arrays associated to the DFL_BLOCKS grid\n");
      printf("-----------------------------------------------------------\n\n");

      printf("%dx%dx%dx%d lattice, ",NPROC0*L0,NPROC1*L1,NPROC2*L2,NPROC3*L3);
      printf("%dx%dx%dx%d process grid, ",NPROC0,NPROC1,NPROC2,NPROC3);
      printf("%dx%dx%dx%d local lattice\n\n",L0,L1,L2,L3);

      read_line("bs","%d %d %d %d",&bs[0],&bs[1],&bs[2],&bs[3]);
      fclose(fin);

      printf("bs = %d %d %d %d\n\n",bs[0],bs[1],bs[2],bs[3]);
      fflush(flog);
   }

   MPI_Bcast(bs,4,MPI_INT,0,MPI_COMM_WORLD);
   
   geometry();

   set_dfl_parms(bs,4,4,1,1,1,1,1,1,0.12,1.0e-3,1.0e-6,1.0e-12,0,0);
   
   inn=vinn(&nb);
   ipp=vipp(nbbe,nbbo,ofsbbe,ofsbbo);
   b=blk_list(DFL_BLOCKS,&nbs,&isw);

   error((bs[0]!=(*b).bs[0])||(bs[1]!=(*b).bs[1])||(bs[2]!=(*b).bs[2])||
         (bs[3]!=(*b).bs[3]),1,"main [check1.c]",
         "Block size bs is incorrectly set"); 
   
   error((nb!=nbs)||(nb<2)||((nb%2)!=0),1,"main [check1.c]",
         "Incorrect block number nb");

   ie=0;
   nbb=nbbe[0]+nbbo[0];

   if (ofsbbe[0]!=0)
      ie=1;
   if (ofsbbo[0]!=(ofsbbe[7]+nbbe[7]))
      ie=2;

   for (ifc=1;ifc<8;ifc++)
   {
      nbb+=(nbbe[ifc]+nbbo[ifc]);

      if (ofsbbe[ifc]!=(ofsbbe[ifc-1]+nbbe[ifc-1]))
         ie=1;
      if (ofsbbo[ifc]!=(ofsbbo[ifc-1]+nbbo[ifc-1]))
         ie=2;
   }

   error(ie==1,1,"main [check1.c]","Incorrect offsets offbbe[ifc]");  
   error(ie==2,1,"main [check1.c]","Incorrect offsets offbbo[ifc]");

   iz=0;
   
   for (ifc=0;ifc<8;ifc++)
   {
      for (ix=ofsbbe[ifc];ix<(ofsbbe[ifc]+nbbe[ifc]);ix++)
      {
         iy=ipp[ix];

         if ((ix>ofsbbe[ifc])&&(iy<=iz))
            ie=1;

         if (inn[iy][ifc]!=(nb+ix))
            ie=3;         
         
         iz=iy;
      }
      
      for (ix=ofsbbo[ifc];ix<(ofsbbo[ifc]+nbbo[ifc]);ix++)
      {
         iy=ipp[ix];

         if ((ix>ofsbbo[ifc])&&(iy<=iz))
            ie=2;

         if (inn[iy][ifc]!=(nb+ix))
            ie=3;         
         
         iz=iy;
      }
   }   
   
   error(ie==1,1,"main [check1.c]","Incorrect ipp at even boundary points");
   error(ie==2,1,"main [check1.c]","Incorrect ipp at odd boundary points");
   error(ie==3,1,"main [check1.c]","ipp and inn are inconsistent");   

   for (ix=0;ix<nb;ix++)
   {
      for (ifc=0;ifc<8;ifc++)
      {
         iy=inn[ix][ifc];

         if ((iy<0)||(iy>=(nb+nbb)))
            ie=1;
         else
         {
            if (iy<nb)
               iz=inn[iy][ifc^0x1];
            else
               iz=ipp[iy-nb];

            if (iz!=ix)
               ie=2;
         }
      }
   }

   error(ie==1,1,"main [check1.c]","Index inn[ix][ifc] is out of range");
   error(ie==2,1,"main [check1.c]","Neighbouring blocks are not paired");   

   l[0]=L0;
   l[1]=L1;
   l[2]=L2;
   l[3]=L3;   
   
   for (ix=0;ix<nb;ix++)
   {
      if (isw==0)
         bo1=b[ix].bo;
      else
         bo1=b[(ix+(nb/2))%nb].bo;
      
      for (ifc=0;ifc<8;ifc++)
      {
         iy=inn[ix][ifc];

         if (iy<nb)
         {
            if (isw==0)
               bo2=b[iy].bo;
            else
               bo2=b[(iy+(nb/2))%nb].bo;         

            for (mu=0;mu<4;mu++)
            {
               if (mu!=(ifc/2))
               {
                  if (bo2[mu]!=bo1[mu])
                     ie=1;
               }
               else
               {
                  is=2*(ifc&0x1)-1;
                  is=(bo1[mu]+is*bs[mu]+l[mu])%l[mu];
                  
                  if (bo2[mu]!=is)
                     ie=1;
               }
            }
         }
      }
   }

   error(ie==1,1,"main [check1.c]","Block and index ordering differ");
   
   if (my_rank==0)
   {
      printf("The block lattice is correctly mapped by the index arrays\n\n");
      fclose(flog);
   }   

   MPI_Finalize(); 
   exit(0);
}
