
/*******************************************************************************
*
* File check12.c
*
* Copyright (C) 2007 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Propagation of the deflation subspace
*
*******************************************************************************/

#define MAIN_PROGRAM

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "mpi.h"
#include "random.h"
#include "start.h"
#include "misc.h"
#include "flags.h"
#include "block.h"
#include "linalg.h"
#include "linsolv.h"
#include "sw_term.h"
#include "dirac.h"
#include "sap_gcr.h"
#include "hmc.h"
#include "dfl.h"
#include "global.h"

static complex **vs;


static void update_ud(double eps)
{
   int ifc;
   su3_alg_dble X[8],*Y;
   su3_dble *u,*um;
   
   u=pud[VOLUME/2][0];
   um=u+4*VOLUME;

   for (;u<um;)
   {
      random_alg(8,X);
      Y=X;

      for (ifc=0;ifc<8;ifc++)
      {
         ExpX(eps,Y,u);
         Y+=1;
         u+=1;
      }
   }

   set_flags(NEW_UD);

   assign_ud2u();   
   sw_term();
   assign_swd2sw();
}


static void check_ortho(double *dev1,double *dev2)
{
   int Ns,nb,isw,nv,i,j;
   complex z;
   double d1,d2;
   dfl_parms_t dfl;

   dfl=dfl_parms();
   Ns=dfl.Ns;

   if (vs==NULL)
   {
      vs=amalloc(Ns*sizeof(*vs),3);
      error(vs==NULL,1,"check_ortho [check12.c]",
            "Unable to allocate auxiliary array");
   }

   blk_list(DFL_BLOCKS,&nb,&isw);
   nv=nb*Ns;
   
   dfl_global_modes(vs);

   (*dev1)=0.0;
   (*dev2)=0.0;

   for (i=0;i<Ns;i++)
   {
      d1=fabs(1.0-(double)(vnorm_square(nv,1,vs[i])));

      if (d1>(*dev1))
         (*dev1)=d1;

      for (j=0;j<i;j++)
      {
         z=vprod(nv,1,vs[i],vs[j]);
         d2=sqrt((double)(z.re*z.re+z.im*z.im));

         if (d2>(*dev2))
            (*dev2)=d2;         
      }
   }
}



int main(int argc,char *argv[])
{
   int my_rank,id;
   int bc,nc_first,nc_last,nc_step;
   int bs_sap[4],nmr_sap,ncy_sap,nkv_sap,nmx;
   int bs_dfl[4],Ns,rinv,rnmr,rncy,rvnmx;
   int vnkv,vnmx,vdnmx;
   int nsize,icnfg,n,iwd0,iwd1;
   int stat_dfl,stat_upd,status[3],ifail;
   double beta,kappa,csw,eps,res;
   double rkappa,rvres,vres,vdres;
   double dev1,dev2;
   spinor_dble **wsd;
   char cnfg_dir[NAME_SIZE],cnfg_file[NAME_SIZE];
   char nbase[NAME_SIZE];   
   lat_parms_t lat;
   gcr_parms_t gcr;   
   dfl_parms_t dfl;
   FILE *fin=NULL,*flog=NULL;   

   MPI_Init(&argc,&argv);
   MPI_Comm_rank(MPI_COMM_WORLD,&my_rank);

   if (my_rank==0)
   {
      flog=freopen("check12.log","w",stdout);
      fin=freopen("check12.in","r",stdin);
      
      printf("\n");
      printf("Propagation of the deflation subspace\n");
      printf("-------------------------------------\n\n");

      printf("%dx%dx%dx%d lattice, ",NPROC0*L0,NPROC1*L1,NPROC2*L2,NPROC3*L3);
      printf("%dx%dx%dx%d process grid, ",NPROC0,NPROC1,NPROC2,NPROC3);
      printf("%dx%dx%dx%d local lattice\n\n",L0,L1,L2,L3);

      read_line("id","%d\n",&id);
      read_line("cnfg_dir","%s\n",cnfg_dir);
      read_line("beta","%lf\n",&beta);
      read_line("kappa","%lf\n",&kappa);
      read_line("csw","%lf\n",&csw);      
      read_line("bc","%d\n",&bc);      
      read_line("nc_first","%d\n",&nc_first);
      read_line("nc_last","%d\n",&nc_last);      
      read_line("nc_step","%d\n",&nc_step);      
      
      read_line("bs_sap","%d %d %d %d\n",
                &bs_sap[0],&bs_sap[1],&bs_sap[2],&bs_sap[3]);
      read_line("nmr_sap","%d\n",&nmr_sap);   
      read_line("ncy_sap","%d\n",&ncy_sap);
      read_line("nkv_sap","%d\n",&nkv_sap);         
      read_line("nmx","%d\n",&nmx);      
      read_line("res","%lf\n",&res);      

      read_line("bs_dfl","%d %d %d %d\n",
                &bs_dfl[0],&bs_dfl[1],&bs_dfl[2],&bs_dfl[3]);
      read_line("Ns","%d\n",&Ns);
      read_line("rkappa","%lf\n",&rkappa);
      read_line("rinv","%d\n",&rinv);
      read_line("rnmr","%d\n",&rnmr);
      read_line("rncy","%d\n",&rncy);
      read_line("rvnmx","%d\n",&rvnmx);      
      read_line("rvres","%lf\n",&rvres);
      read_line("vnkv","%d\n",&vnkv);
      read_line("vnmx","%d\n",&vnmx);       
      read_line("vres","%lf\n",&vres);
      read_line("vdnmx","%d\n",&vdnmx);       
      read_line("vdres","%lf\n",&vdres);
      read_line("eps","%lf",&eps);
      fclose(fin);
   }

   MPI_Bcast(&id,1,MPI_INT,0,MPI_COMM_WORLD);      
   MPI_Bcast(cnfg_dir,NAME_SIZE,MPI_CHAR,0,MPI_COMM_WORLD);
   MPI_Bcast(&beta,1,MPI_DOUBLE,0,MPI_COMM_WORLD);
   MPI_Bcast(&kappa,1,MPI_DOUBLE,0,MPI_COMM_WORLD);
   MPI_Bcast(&csw,1,MPI_DOUBLE,0,MPI_COMM_WORLD);
   MPI_Bcast(&bc,1,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&nc_first,1,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&nc_last,1,MPI_INT,0,MPI_COMM_WORLD);   
   MPI_Bcast(&nc_step,1,MPI_INT,0,MPI_COMM_WORLD);   
   
   MPI_Bcast(bs_sap,4,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&nmr_sap,1,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&ncy_sap,1,MPI_INT,0,MPI_COMM_WORLD);   
   MPI_Bcast(&nkv_sap,1,MPI_INT,0,MPI_COMM_WORLD);   
   MPI_Bcast(&nmx,1,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&res,1,MPI_DOUBLE,0,MPI_COMM_WORLD);

   MPI_Bcast(bs_dfl,4,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&Ns,1,MPI_INT,0,MPI_COMM_WORLD);   
   MPI_Bcast(&rkappa,1,MPI_DOUBLE,0,MPI_COMM_WORLD);
   MPI_Bcast(&rinv,1,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&rnmr,1,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&rncy,1,MPI_INT,0,MPI_COMM_WORLD);   
   MPI_Bcast(&rvnmx,1,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&rvres,1,MPI_DOUBLE,0,MPI_COMM_WORLD);
   MPI_Bcast(&vnkv,1,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&vnmx,1,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&vres,1,MPI_DOUBLE,0,MPI_COMM_WORLD);
   MPI_Bcast(&vdnmx,1,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&vdres,1,MPI_DOUBLE,0,MPI_COMM_WORLD);
   MPI_Bcast(&eps,1,MPI_DOUBLE,0,MPI_COMM_WORLD);
   
   start_ranlux(0,123456);   
   geometry();

   alloc_ud();
   alloc_swd();
   alloc_sd(5);
   
   alloc_u();   
   alloc_sw();
   
   if (Ns<(2*nkv_sap))
      alloc_s(2*nkv_sap+1);
   else
      alloc_s(Ns+2);
   
   lat=set_lat_parms(beta,kappa,csw);   
   gcr=set_gcr_parms(bs_sap,nmr_sap,ncy_sap,nkv_sap);
   dfl=set_dfl_parms(bs_dfl,Ns,rinv,rnmr,rncy,rvnmx,vnkv,vnmx,vdnmx,
                     rkappa,rvres,vres,vdres,0,0);

   if (my_rank==0)
   {
      printf("beta = %.6f\n",lat.beta);
      printf("kappa = %.6f\n",lat.kappa);
      printf("csw = %.6f\n",lat.csw);
      printf("bc = %+d\n\n",bc);

      printf("bs_sap = %d %d %d %d\n",
             bs_sap[0],bs_sap[1],bs_sap[2],bs_sap[3]);
      printf("nmr = %d\n",gcr.nmr);       
      printf("ncy = %d\n",gcr.ncy);
      printf("nkv = %d\n",gcr.nkv);
      printf("nmx = %d\n",nmx);
      printf("res = %.1e\n\n",res);
      
      printf("bs_dfl = %d %d %d %d\n",
             bs_dfl[0],bs_dfl[1],bs_dfl[2],bs_dfl[3]);
      printf("Ns = %d\n",dfl.Ns);
      printf("rkappa = %.6f\n",dfl.rkappa);       
      printf("rinv = %d\n",dfl.rinv);
      printf("rnmr = %d\n",dfl.rnmr);       
      printf("rncy = %d\n",dfl.rncy);
      printf("rvnmx = %d\n",dfl.rvnmx);
      printf("rvres = %.1e\n",dfl.rvres);
      printf("vnkv = %d\n",dfl.vnkv);
      printf("vnmx = %d\n",dfl.vnmx);
      printf("vres = %.1e\n",dfl.vres);
      printf("vdnmx = %d\n",dfl.vdnmx);
      printf("vdres = %.1e\n",dfl.vdres);
      printf("eps = %.1e\n\n",eps);

      printf("Configurations no %d -> %d in steps of %d\n\n",
             nc_first,nc_last,nc_step);      
      fflush(flog);      
   }
   
   wsd=reserve_wsd(2);
   iwd0=(wsd[0]-psd[0][0])/NSPIN;
   iwd1=(wsd[1]-psd[0][0])/NSPIN;
   random_sd(VOLUME,wsd[0],1.0);
   
   check_dir_root(cnfg_dir);   

   nsize=name_size("%dx%dx%dx%db%.2fk%.5fc%.5fid%d",
                   NPROC0*L0,NPROC1*L1,NPROC2*L2,NPROC3*L3,beta,kappa,csw,id);
   error_root(nsize>=NAME_SIZE,1,"main [check12.c]","Base name is too long");
   sprintf(nbase,"%dx%dx%dx%db%.2fk%.5fc%.5fid%d",
           NPROC0*L0,NPROC1*L1,NPROC2*L2,NPROC3*L3,beta,kappa,csw,id);

   nsize=name_size("%s/%sn%d",cnfg_dir,nbase,nc_last);
   error_root(nsize>=NAME_SIZE,1,"main [check12.c]",
              "cnfg_dir name is too long");

   for (icnfg=nc_first;icnfg<=nc_last;icnfg+=nc_step)
   {
      if (my_rank==0)
      {
         printf("Configuration no %d:\n\n",icnfg);
         fflush(flog);
      }

      sprintf(cnfg_file,"%s/%sn%d",cnfg_dir,nbase,icnfg);
      import_cnfg(cnfg_file);
      sw_term();
      if (bc==-1)
         flipbcd();

      dfl_modes(1,&stat_dfl,&ifail);
      check_ortho(&dev1,&dev2);
      
      if (my_rank==0)
      {
         printf("Subspace initialization:\n");
         printf("stat_dfl = %d, ifail = %d\n",stat_dfl,ifail);
         printf("Orthonormality = %.1e (norm), %.1e (scalar products)\n",
                dev1,dev2);
      }

      error((stat_dfl<0)||(ifail!=0),1,
            "main [check12.c]","dfl_modes() failed");
      
      dfl_sap_gcr(nmx,res,iwd0,iwd1,status,&ifail);

      if (my_rank==0)
      {
         printf("Solution of the Dirac equation:\n");
         printf("status = %d (little %d, %d), ifail = %d\n\n",
                status[0],status[1],status[2],ifail);
      }
      
      error((status[0]<0)||(status[2]<0)||(ifail!=0),1,
            "main [check12.c]","dfl_sap_gcr() failed");

      for (n=0;n<5;n++)
      {
         if (my_rank==0)
            printf("Field update no %d:\n",n+1);

         update_ud(eps);

         dfl_update(n&0x1,1,&stat_upd,&ifail);
         check_ortho(&dev1,&dev2);
         
         if (my_rank==0)
         {
            printf("Deflation subspace updated:\n");
            printf("status = %d, ifail = %d\n",stat_upd,ifail);
            printf("Orthonormality = %.1e (norm), %.1e (scalar products)\n",
                   dev1,dev2);
         }

         error((stat_upd<0)||(ifail!=0),1,
               "main [check12.c]","dfl_update() failed");
         
         dfl_sap_gcr(nmx,res,iwd0,iwd1,status,&ifail);

         if (my_rank==0)
         {
            printf("Solution of the Dirac equation:\n");
            printf("status = %d (little %d, %d), ifail = %d\n\n",
                   status[0],status[1],status[2],ifail);
         }
      
         error((status[0]<0)||(status[2]<0)||(ifail!=0),1,
               "main [check12.c]","dfl_sap_gcr() failed");
      }

      error_chk();
      
      if (my_rank==0)
         fflush(flog);
   }

   release_wsd();
   error_chk();
   
   if (my_rank==0)
   {
      printf("\n");
      fclose(flog);
   }
   
   MPI_Finalize();   
   exit(0);
}
