
/*******************************************************************************
*
* File check2.c
*
* Copyright (C) 2007 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Allocation and initialization of the global vector fields
*
*******************************************************************************/

#define MAIN_PROGRAM

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "mpi.h"
#include "random.h"
#include "start.h"
#include "misc.h"
#include "dfl.h"
#include "global.h"
   
#define NFIELDS 8

static float sig[NFIELDS];
static double sigd[NFIELDS];


int main(int argc,char *argv[])
{
   int my_rank,bs[4],Ns,vnkv;
   int nb,nbb,n,nv,nvec;
   int ie,k,i,ix;
   float d,dmax,dmax_all;
   double var,var_all;
   complex *ss,**wv,*v;
   complex_dble **wvd,*vd;
   dfl_parms_t dfl;
   FILE *fin=NULL,*flog=NULL;   

   MPI_Init(&argc,&argv);
   MPI_Comm_rank(MPI_COMM_WORLD,&my_rank);

   if (my_rank==0)
   {
      flog=freopen("check2.log","w",stdout);
      fin=freopen("check2.in","r",stdin);
      
      printf("\n");
      printf("Allocation and initialization of the global vector fields\n");
      printf("---------------------------------------------------------\n\n");

      printf("%dx%dx%dx%d lattice, ",NPROC0*L0,NPROC1*L1,NPROC2*L2,NPROC3*L3);
      printf("%dx%dx%dx%d process grid, ",NPROC0,NPROC1,NPROC2,NPROC3);
      printf("%dx%dx%dx%d local lattice\n\n",L0,L1,L2,L3);

      read_line("bs","%d %d %d %d",&bs[0],&bs[1],&bs[2],&bs[3]);
      read_line("Ns","%d",&Ns);
      read_line("vnkv","%d",&vnkv);
      fclose(fin);

      printf("bs = %d %d %d %d\n",bs[0],bs[1],bs[2],bs[3]);
      printf("Ns = %d\n",Ns);
      printf("vnkv = %d\n\n",vnkv);
      fflush(flog);
   }

   MPI_Bcast(bs,4,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&Ns,1,MPI_INT,0,MPI_COMM_WORLD);   
   MPI_Bcast(&vnkv,1,MPI_INT,0,MPI_COMM_WORLD);
   
   start_ranlux(0,123456);   
   geometry();

   dfl=set_dfl_parms(bs,Ns,4,1,1,1,vnkv,1,1,0.12,1.0e-3,1.0e-6,1.0e-12,
                     2*NFIELDS,NFIELDS);   

   error((Ns!=dfl.Ns)||(vnkv!=dfl.vnkv),1,"main [check2.c]",
         "Parameter Ns or vnkv is incorrectly set");
   
   wv=reserve_wv(NFIELDS);
   wvd=reserve_wvd(NFIELDS);   

   nb=VOLUME/(bs[0]*bs[1]*bs[2]*bs[3]);
   nbb=(FACE0/(bs[1]*bs[2]*bs[3])+
        FACE1/(bs[0]*bs[2]*bs[3])+
        FACE2/(bs[0]*bs[1]*bs[3])+
        FACE3/(bs[0]*bs[1]*bs[2]));      
   
   v=vflds_base(&n,&nv,&nvec);
   error((v!=wv[0])||(n!=(2*vnkv+2+2*NFIELDS))||(nv!=(nb*Ns))||
         (nvec!=((nb+nbb))*Ns),1,"main [check2.c]",
         "vbase() returns incorrect values");

   vd=vdflds_base(&n,&nv,&nvec);
   error((vd!=wvd[0])||(n!=(4+NFIELDS))||(nv!=(nb*Ns))||
         (nvec!=((nb+nbb)*Ns)),1,"main [check2.c]",
         "vdbase() returns incorrect values");   

   ie=0;

   for (k=0;k<NFIELDS;k++)
   {
      if ((wv[k]-v)!=(k*nvec))
         ie=1;
      if ((wvd[k]-vd)!=(k*nvec))
         ie=2;      
   }

   error(ie==1,1,"main [check2.c]",
         "Field addresses reserved by wv_reserve() are incorrect");
   error(ie==2,1,"main [check2.c]",
         "Field addresses reserved by wvd_reserve() are incorrect");

   dmax=0.0f;

   for (ix=0;ix<nvec;ix++)
   {
      for (k=0;k<NFIELDS;k++)
      {
         d=(float)(fabs((double)(wv[k][ix].re))+
                   fabs((double)(wv[k][ix].im)));

         if (d>dmax)
            dmax=d;
      }
   }

   MPI_Reduce(&dmax,&dmax_all,1,MPI_FLOAT,MPI_MAX,0,MPI_COMM_WORLD);
   
   if (my_rank==0)
   {
      printf("Allocate %d single-precision vector fields\n",NFIELDS);
      printf("|v| = %.2e\n\n",dmax_all);

      printf("Choose random single-precision fields\n");   
   }

   if (my_rank==0)
      ranlxs(sig,NFIELDS);

   MPI_Bcast(sig,NFIELDS,MPI_FLOAT,0,MPI_COMM_WORLD);

   for (k=0;k<NFIELDS;k++)
   {
      random_v(nv,wv[k],sig[k]);
      var=0.0;

      for (ix=0;ix<nv;ix++)
         var+=(double)(wv[k][ix].re*wv[k][ix].re+
                       wv[k][ix].im*wv[k][ix].im);

      MPI_Reduce(&var,&var_all,1,MPI_DOUBLE,MPI_SUM,0,MPI_COMM_WORLD);
      
      if (my_rank==0)
      {
         var_all/=(double)(nv*NPROC);               
         printf("<|v[%d]|^2> = %.4e (sigma^2 = %.4e)\n",
                k,var_all,sig[k]*sig[k]);
      }
   }

   dmax=0.0f;

   for (ix=0;ix<nvec;ix++)
   {
      for (k=0;k<NFIELDS;k++)
      {
         d=(float)(fabs(wvd[k][ix].re)+
                   fabs(wvd[k][ix].im));
         
         if (d>dmax)
            dmax=d;
      }
   }

   MPI_Reduce(&dmax,&dmax_all,1,MPI_FLOAT,MPI_MAX,0,MPI_COMM_WORLD);
   
   if (my_rank==0)
   {   
      printf("\n");
      printf("Allocate %d double-precision vector fields\n",NFIELDS);
      printf("|v| = %.2e\n\n",dmax_all);

      printf("Choose random double-precision vector fields\n");   
   }
   
   if (my_rank==0)
      ranlxd(sigd,NFIELDS);

   MPI_Bcast(sigd,NFIELDS,MPI_DOUBLE,0,MPI_COMM_WORLD);
   
   for (k=0;k<NFIELDS;k++)
   {
      random_vd(nv,wvd[k],sigd[k]);
      var=0.0;

      for (ix=0;ix<nv;ix++)
         var+=(wvd[k][ix].re*wvd[k][ix].re+
               wvd[k][ix].im*wvd[k][ix].im);

      MPI_Reduce(&var,&var_all,1,MPI_DOUBLE,MPI_SUM,0,MPI_COMM_WORLD);
      
      if (my_rank==0)
      {
         var_all/=(double)(nv*NPROC);               
         printf("<|v[%d]|^2> = %.4e (sigma^2 = %.4e)\n",
                k,var_all,sigd[k]*sigd[k]);
      }
   }

   ss=malloc(NFIELDS*nv*sizeof(complex));
   i=0;

   for (k=0;k<NFIELDS;k++)
   {
      for (ix=0;ix<nv;ix++)
      {
         ss[i].re=wv[k][ix].re;
         ss[i].im=wv[k][ix].im;
         i+=1;
      }
   }   

   for (k=0;k<NFIELDS;k++)
      assign_v2vd(nv,wv[k],wvd[NFIELDS-1-k]);

   release_wv();
   free_vflds();
   alloc_vflds(NFIELDS);
   wv=reserve_wv(NFIELDS);

   for (k=0;k<NFIELDS;k++)
      assign_vd2v(nv,wvd[k],wv[NFIELDS-1-k]);   

   i=0;
   dmax=0.0f;

   for (k=0;k<NFIELDS;k++)
   {
      for (ix=0;ix<nv;ix++)
      {
         d=(float)(fabs((double)(wv[k][ix].re-ss[i].re))+
                   fabs((double)(wv[k][ix].im-ss[i].im)));
         i+=1;

         if (d>dmax)
            dmax=d;
      }
   }    

   MPI_Reduce(&dmax,&dmax_all,1,MPI_FLOAT,MPI_MAX,0,MPI_COMM_WORLD);
   
   if (my_rank==0)
   { 
      printf("\n");
      printf("Assign v to vd, free v, reassign v from vd\n");
      printf("|v_original-v_assigned| = %.2e\n\n",dmax_all);
   }

   release_wv();
   alloc_vflds(2*NFIELDS);
   wv=reserve_wv(2*NFIELDS);
   
   for (k=0;k<NFIELDS;k++)
   {
      random_v(nv,wv[k],1.0f);
      assign_v2v(nv,wv[k],wv[2*NFIELDS-1-k]);
   }

   dmax=0.0f;

   for (k=0;k<NFIELDS;k++)
   {
      for (ix=0;ix<nv;ix++)
      {
         d=(float)(fabs((double)(wv[2*NFIELDS-1-k][ix].re-wv[k][ix].re))+
                   fabs((double)(wv[2*NFIELDS-1-k][ix].im-wv[k][ix].im)));

         if (d>dmax)
            dmax=d;
      }
   }    

   MPI_Reduce(&dmax,&dmax_all,1,MPI_FLOAT,MPI_MAX,0,MPI_COMM_WORLD);
   
   if (my_rank==0)
   { 
      printf("Assign v to v\n");
      printf("|v_original-v_assigned| = %.2e\n\n",dmax_all);
   }

   release_wvd();
   alloc_vdflds(2*NFIELDS);
   wvd=reserve_wvd(2*NFIELDS);
   
   for (k=0;k<NFIELDS;k++)
   {
      random_vd(nv,wvd[k],1.0);
      assign_vd2vd(nv,wvd[k],wvd[2*NFIELDS-1-k]);
   }

   dmax=0.0f;

   for (k=0;k<NFIELDS;k++)
   {
      for (ix=0;ix<nv;ix++)
      {
         d=(float)(fabs(wvd[2*NFIELDS-1-k][ix].re-wvd[k][ix].re)+
                   fabs(wvd[2*NFIELDS-1-k][ix].im-wvd[k][ix].im));
         
         if (d>dmax)
            dmax=d;
      }
   }    

   MPI_Reduce(&dmax,&dmax_all,1,MPI_FLOAT,MPI_MAX,0,MPI_COMM_WORLD);
   
   if (my_rank==0)
   { 
      printf("Assign vd to vd\n");
      printf("|v_original-v_assigned| = %.2e\n\n",dmax_all);

      fclose(flog);
   }
   
   MPI_Finalize();   
   exit(0);
}
