
/*******************************************************************************
*
* File check10.c
*
* Copyright (C) 2005 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Comparision of Qbnd_dble, Qbnds_dble ... with Qnohat_dble
*
*******************************************************************************/

#define MAIN_PROGRAM

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "mpi.h"
#include "su3.h"
#include "random.h"
#include "start.h"
#include "linalg.h"
#include "block.h"
#include "sw_term.h"
#include "dirac.h"
#include "global.h"


static int eoind(int *bs)
{
   int n[4];
   int mu;

   n[0]=(cpr[0]*L0)/bs[0];
   n[1]=(cpr[1]*L1)/bs[1];
   n[2]=(cpr[2]*L2)/bs[2];
   n[3]=(cpr[3]*L3)/bs[3];

   for (mu=0;mu<4;mu++)
   {
      if ((n[mu]&0x1)==1)
         return 1;
   }

   return 0;
}


int main(int argc,char *argv[])
{
   int my_rank,nb,isw,n0;
   int bs[4],vol;
   double d,d1,d2,d3,d4;
   complex_dble z;
   spinor_dble *ps0,*ps1,*ps2;
   spinor_dble *psb0,*psb1,*psb2;
   weyl_dble *psbb0,*psbb1;
   block_t *b0,*b;
   bndry_t *bb;
   FILE *flog=NULL,*fin=NULL;

   MPI_Init(&argc,&argv);
   MPI_Comm_rank(MPI_COMM_WORLD,&my_rank);

   if (my_rank==0)
   {
      flog=freopen("check10.log","w",stdout);
      fin=freopen("check10.in","r",stdin);

      printf("\n");
      printf("Comparision of Qbnd_dble and Qbnds_dble ... with Qnohat_dble\n");
      printf("------------------------------------------------------------\n");
      printf("\n");

      printf("%dx%dx%dx%d lattice, ",NPROC0*L0,NPROC1*L1,NPROC2*L2,NPROC3*L3);
      printf("%dx%dx%dx%d process grid, ",NPROC0,NPROC1,NPROC2,NPROC3);
      printf("%dx%dx%dx%d local lattice\n\n",L0,L1,L2,L3);

      scanf("bs %d %d %d %d",&bs[0],&bs[1],&bs[2],&bs[3]);
      fclose(fin);

      printf("bs = %d %d %d %d\n\n",bs[0],bs[1],bs[2],bs[3]);
   }

   MPI_Bcast(bs,4,MPI_INT,0,MPI_COMM_WORLD);

   start_ranlux(0,1234);
   geometry();
   alloc_ud();
   alloc_sd(3);
   alloc_swd();

   random_ud();
   set_lat_parms(0.0,0.1,0.3);
   sw_term();

   alloc_bgr(HMC_BLOCKS,bs);
   b0=blk_list(HMC_BLOCKS,&nb,&isw);
   isw=eoind(bs);

   for (n0=0;n0<nb;n0++)
   {
      b=b0+n0;
      message("Block number: %d, share flags: %d [%#x], ",
              n0,(*b).shf,(*b).shf);
      (*b).shf&=(~0x20);
      alloc_bnd(b,0,2);
      (*b).shf|=0x20;
      message("%d [%#x] (with new boundary)\n",
              (*b).shf,(*b).shf);
   }

   assign_ud2udbgr(HMC_BLOCKS);
   assign_swd2swdbgr(HMC_BLOCKS,NO_PTS);

   ps0=psd[0][0];
   ps1=psd[1][0];
   ps2=psd[2][0];

   d1=0.0;
   d2=0.0;
   d3=0.0;
   d4=0.0;
   z.im=0.0;

   for (n0=0;n0<nb;n0++)
   {
      b=b0+n0;
      bb=(*b).bb;
      vol=(*b).vbb/2;

      psb0=(*b).sd[0];
      psb1=(*b).sd[1];
      psb2=(*b).sd[2];
      psbb0=(*bb).sd[0];
      psbb1=(*bb).sd[1];

      set_sd2zero(VOLUME,ps0);
      random_sd((*b).vol,psb0,1.0);
      if (isw!=0)
         set_sd2zero((*b).vol,psb0);

      assign_sdb2sd(b,ALL_PTS,0,0);
      Qbnds_dble(b,0,0);
      Qnohat_dble(0,1);
      assign_sd2sdbb(b,1,1);
      if (isw!=0)
         set_sd2zero(vol,(spinor_dble*)(psbb1));

      z.re=-1.0;
      mulc_spinor_add_dble(vol,(spinor_dble*)(psbb0),
                           (spinor_dble*)(psbb1),z);
      d=norm_square_dble(vol,1,(spinor_dble*)(psbb0))/
            norm_square_dble(vol,1,(spinor_dble*)(psbb1));
      if (d>d1)
         d1=d;

      random_sd(VOLUME,ps0,1.0);
      assign_sd2sdb(b,ALL_PTS,0,0);
      set_sd2zero((*b).vol,psb1);
      if (isw==0)
         assign_sdb2sd(b,ALL_PTS,1,0);
      assign_sd2sd((*b).vol,psb0,psb1);

      assign_sd2sdbb(b,0,0);
      Qbnd_dble(b,0,0);
      Qnohat_dble(0,1);
      assign_sd2sdb(b,ALL_PTS,1,2);
      z.re=-1.0;
      mulc_spinor_add_dble((*b).vol,psb1,psb2,z);
      mulc_spinor_add_dble((*b).vol,psb0,psb1,z);

      if (isw!=0)
         set_sd2zero((*b).vol,psb0);
      d=norm_square_dble((*b).vol,1,psb0)/
            norm_square_dble((*b).vol,1,psb1);
      if (d>d2)
         d2=d;

      random_sd((*b).vol,psb0,1.0);
      if (isw!=0)
         set_sd2zero((*b).vol,psb0);
      set_sd2zero(VOLUME,ps0);
      assign_sdb2sd(b,ALL_PTS,0,0);
      Qnohat_dble(0,1);
      set_sd2zero((*b).vol,psb1);
      if (isw==0)
         assign_sdb2sd(b,ALL_PTS,1,1);

      random_sd(VOLUME,ps0,1.0);
      assign_sd2sd(VOLUME,ps0,ps2);
      Qbnds_dble(b,0,0);
      sub_assign_sdbb2sd(b,0,0);

      z.re=1.0;
      mulc_spinor_add_dble(VOLUME,ps0,ps1,z);
      z.re=-1.0;
      mulc_spinor_add_dble(VOLUME,ps0,ps2,z);
      d=norm_square_dble(VOLUME,1,ps0)/
            norm_square_dble(VOLUME,1,ps2);
      if (d>d3)
         d3=d;

      random_sd((*b).vol,psb0,1.0);
      if (isw!=0)
         set_sd2zero((*b).vol,psb0);
      set_sd2zero(VOLUME,ps0);
      assign_sdb2sd(b,ALL_PTS,0,0);
      Qnohat_dble(0,1);
      mulg5_dble(VOLUME,ps1);
      set_sd2zero((*b).vol,psb1);
      if (isw==0)
         assign_sdb2sd(b,ALL_PTS,1,1);

      random_sd(VOLUME,ps0,1.0);
      assign_sd2sd(VOLUME,ps0,ps2);
      Qbnds_dble(b,0,0);
      mulg5_sub_assign_sdbb2sd(b,0,0);

      z.re=1.0;
      mulc_spinor_add_dble(VOLUME,ps0,ps1,z);
      z.re=-1.0;
      mulc_spinor_add_dble(VOLUME,ps0,ps2,z);
      d=norm_square_dble(VOLUME,1,ps0)/
            norm_square_dble(VOLUME,1,ps2);
      if (d>d4)
         d4=d;
   }

   free_bgr(HMC_BLOCKS);
   error_chk();

   if (my_rank==0)
   {
      printf("\n");
      printf("Check of Qbnds_dble and assign_sd2sdbb: %.2e\n\n",
             sqrt(d1));
      printf("Check of Qbnd_dble and assign_sd2sdbb: %.2e\n\n",
             sqrt(d2));
      printf("Check of Qbnds_dble followed by sub_assign_sd2sdbb: %.2e\n\n",
             sqrt(d3));
      printf("Check of Qbnds_dble followed by ");
      printf("mulg5_sub_assign_sd2sdbb: %.2e\n\n",
             sqrt(d4));

      fclose(flog);
   }

   MPI_Finalize();
   exit(0);
}
