
/*******************************************************************************
*
* File check9.c
*
* Copyright (C) 2005 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Comparision of Qbnd, Qbnds ... with Qnohat
*
*******************************************************************************/

#define MAIN_PROGRAM

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "mpi.h"
#include "su3.h"
#include "random.h"
#include "start.h"
#include "linalg.h"
#include "block.h"
#include "sw_term.h"
#include "dirac.h"
#include "global.h"


static int eoind(int *bs)
{
   int n[4];
   int mu;

   n[0]=(cpr[0]*L0)/bs[0];
   n[1]=(cpr[1]*L1)/bs[1];
   n[2]=(cpr[2]*L2)/bs[2];
   n[3]=(cpr[3]*L3)/bs[3];

   for (mu=0;mu<4;mu++)
   {
      if ((n[mu]&0x1)==1)
         return 1;
   }

   return 0;
}


int main(int argc,char *argv[])
{
   int my_rank,nb,isw,n0;
   int bs[4],vol;
   float d,d1,d2,d3,d4;
   complex z;
   spinor *ps0,*ps1,*ps2;
   spinor *psb0,*psb1,*psb2;
   weyl *psbb0,*psbb1;
   block_t *b0,*b;
   bndry_t *bb;
   FILE *flog=NULL,*fin=NULL;

   MPI_Init(&argc,&argv);
   MPI_Comm_rank(MPI_COMM_WORLD,&my_rank);

   if (my_rank==0)
   {
      flog=freopen("check9.log","w",stdout);
      fin=freopen("check9.in","r",stdin);

      printf("\n");
      printf("Comparision of Qbnd and Qbnds ... with Qnohat\n");
      printf("---------------------------------------------\n\n");

      printf("%dx%dx%dx%d lattice, ",NPROC0*L0,NPROC1*L1,NPROC2*L2,NPROC3*L3);
      printf("%dx%dx%dx%d process grid, ",NPROC0,NPROC1,NPROC2,NPROC3);
      printf("%dx%dx%dx%d local lattice\n\n",L0,L1,L2,L3);

      scanf("bs %d %d %d %d",&bs[0],&bs[1],&bs[2],&bs[3]);
      fclose(fin);

      printf("bs = %d %d %d %d\n\n",bs[0],bs[1],bs[2],bs[3]);
   }

   MPI_Bcast(bs,4,MPI_INT,0,MPI_COMM_WORLD);

   start_ranlux(0,1234);
   geometry();
   alloc_u();
   alloc_ud();
   alloc_s(3);
   alloc_sw();
   alloc_swd();

   random_ud();
   set_lat_parms(0.0,0.1,0.3);
   sw_term();
   assign_ud2u();
   assign_swd2sw();

   alloc_bgr(GCR_BLOCKS,bs);
   b0=blk_list(GCR_BLOCKS,&nb,&isw);
   isw=eoind(bs);

   for (n0=0;n0<nb;n0++)
   {
      b=b0+n0;
      message("Block number: %d, share flags: %d [%#x], ",
              n0,(*b).shf,(*b).shf);
      (*b).shf&=(~0x20);
      alloc_bnd(b,2,0);
      (*b).shf|=0x20;
      message("%d [%#x] (with new boundary)\n",
              (*b).shf,(*b).shf);
   }

   assign_u2ubgr(GCR_BLOCKS);
   assign_swd2swbgr(GCR_BLOCKS,NO_PTS);

   ps0=ps[0][0];
   ps1=ps[1][0];
   ps2=ps[2][0];

   d1=0.0f;
   d2=0.0f;
   d3=0.0f;
   d4=0.0f;
   z.im=0.0f;   

   for (n0=0;n0<nb;n0++)
   {
      b=b0+n0;
      bb=(*b).bb;
      vol=(*b).vbb/2;

      psb0=(*b).s[0];
      psb1=(*b).s[1];
      psb2=(*b).s[2];
      psbb0=(*bb).s[0];
      psbb1=(*bb).s[1];

      set_s2zero(VOLUME,ps0);
      random_s((*b).vol,psb0,1.0f);
      if (isw!=0)
         set_s2zero((*b).vol,psb0);

      assign_sb2s(b,ALL_PTS,0,0);
      Qbnds(b,0,0);
      Qnohat(0,1);
      assign_s2sbb(b,1,1);
      if (isw!=0)
         set_s2zero(vol,(spinor*)(psbb1));

      z.re=-1.0f;
      mulc_spinor_add(vol,(spinor*)(psbb0),(spinor*)(psbb1),z);
      d=norm_square(vol,1,(spinor*)(psbb0))/
            norm_square(vol,1,(spinor*)(psbb1));
      if (d>d1)
         d1=d;

      random_s(VOLUME,ps0,1.0f);
      assign_s2sb(b,ALL_PTS,0,0);
      set_s2zero((*b).vol,psb1);
      if (isw==0)
         assign_sb2s(b,ALL_PTS,1,0);
      assign_s2s((*b).vol,psb0,psb1);

      assign_s2sbb(b,0,0);
      Qbnd(b,0,0);
      Qnohat(0,1);
      assign_s2sb(b,ALL_PTS,1,2);

      z.re=-1.0f;
      mulc_spinor_add((*b).vol,psb1,psb2,z);
      mulc_spinor_add((*b).vol,psb0,psb1,z);

      if (isw!=0)
         set_s2zero((*b).vol,psb0);
      d=norm_square((*b).vol,1,psb0)/norm_square((*b).vol,1,psb1);
      if (d>d2)
         d2=d;

      random_s((*b).vol,psb0,1.0f);
      if (isw!=0)
         set_s2zero((*b).vol,psb0);
      set_s2zero(VOLUME,ps0);
      assign_sb2s(b,ALL_PTS,0,0);
      Qnohat(0,1);
      set_s2zero((*b).vol,psb1);
      if (isw==0)
         assign_sb2s(b,ALL_PTS,1,1);

      random_s(VOLUME,ps0,1.0f);
      assign_s2s(VOLUME,ps0,ps2);
      Qbnds(b,0,0);
      sub_assign_sbb2s(b,0,0);

      z.re=1.0f;
      mulc_spinor_add(VOLUME,ps0,ps1,z);
      z.re=-1.0f;
      mulc_spinor_add(VOLUME,ps0,ps2,z);
      d=norm_square(VOLUME,1,ps0)/norm_square(VOLUME,1,ps2);
      if (d>d3)
         d3=d;

      random_s((*b).vol,psb0,1.0f);
      if (isw!=0)
         set_s2zero((*b).vol,psb0);
      set_s2zero(VOLUME,ps0);
      assign_sb2s(b,ALL_PTS,0,0);
      Qnohat(0,1);
      mulg5(VOLUME,ps1);
      set_s2zero((*b).vol,psb1);
      if (isw==0)
         assign_sb2s(b,ALL_PTS,1,1);

      random_s(VOLUME,ps0,1.0f);
      assign_s2s(VOLUME,ps0,ps2);
      Qbnds(b,0,0);
      mulg5_sub_assign_sbb2s(b,0,0);

      z.re=1.0f;
      mulc_spinor_add(VOLUME,ps0,ps1,z);
      z.re=-1.0f;
      mulc_spinor_add(VOLUME,ps0,ps2,z);
      d=norm_square(VOLUME,1,ps0)/norm_square(VOLUME,1,ps2);
      if (d>d4)
         d4=d;
   }

   free_bgr(GCR_BLOCKS);
   error_chk();

   if (my_rank==0)
   {
      printf("\n");
      printf("Check of Qbnds and assign_s2sbb: %.2e\n\n",
             sqrt((double)(d1)));
      printf("Check of Qbnd and assign_s2sbb: %.2e\n\n",
             sqrt((double)(d2)));
      printf("Check of Qbnds followed by sub_assign_sbb2s: %.2e\n\n",
             sqrt((double)(d3)));
      printf("Check of Qbnds followed by mulg5_sub_assign_sbb2s: %.2e\n\n",
             sqrt((double)(d4)));
      fclose(flog);
   }

   MPI_Finalize();
   exit(0);
}
