
/*******************************************************************************
*
* File time1b.c
*
* Copyright (C) 2005 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Timing of Qhat_blk
*
*******************************************************************************/

#define MAIN_PROGRAM

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "mpi.h"
#include "su3.h"
#include "flags.h"
#include "random.h"
#include "start.h"
#include "sw_term.h"
#include "dirac.h"
#include "block.h"
#include "global.h"


int main(int argc,char *argv[])
{
   int my_rank,count,n;
   int nb,isw,bs[4];
   double wt1,wt2,wdt;
   block_t *b,*b0,*bm;
   FILE *flog=NULL,*fin=NULL;

   MPI_Init(&argc,&argv);
   MPI_Comm_rank(MPI_COMM_WORLD,&my_rank);

   if (my_rank==0)
   {
      flog=freopen("time1b.log","w",stdout);
      fin=freopen("time1b.in","r",stdin);

      printf("\n");
      printf("Timing of Qhat_blk\n");
      printf("------------------\n\n");

      printf("%dx%dx%dx%d lattice, ",NPROC0*L0,NPROC1*L1,NPROC2*L2,NPROC3*L3);
      printf("%dx%dx%dx%d process grid, ",NPROC0,NPROC1,NPROC2,NPROC3);
      printf("%dx%dx%dx%d local lattice\n\n",L0,L1,L2,L3);

      scanf("bs %d %d %d %d",&bs[0],&bs[1],&bs[2],&bs[3]);
      fclose(fin);

      printf("bs = %d %d %d %d\n\n",bs[0],bs[1],bs[2],bs[3]);
   }

   MPI_Bcast(bs,4,MPI_INT,0,MPI_COMM_WORLD);

   start_ranlux(0,1234);
   geometry();
   alloc_ud();
   alloc_swd();

   random_ud();
   set_lat_parms(0.0,0.1,0.2);
   sw_term();

   alloc_bgr(GCR_BLOCKS,bs);
   assign_ud2ubgr(GCR_BLOCKS);
   error(assign_swd2swbgr(GCR_BLOCKS,ODD_PTS)!=0,1,"main [time1b.c]",
         "Inversion of swd on the odd sites was not safe");
   b0=blk_list(GCR_BLOCKS,&nb,&isw);
   bm=b0+nb;

   random_s((*b0).vol,(*b0).s[0],1.0f);
   random_s((*b0).vol,(*b0).s[1],1.0f);

   n=1000000/VOLUME;
   if (n<1)
      n=1;

   MPI_Barrier(MPI_COMM_WORLD);
   wt1=MPI_Wtime();
   for (count=0;count<n;count++)
   {
      for (b=b0;b<bm;b++)
      {
         Qhat_blk(b,0,1);
         Qhat_blk(b,1,2);
      }
   }
   MPI_Barrier(MPI_COMM_WORLD);
   wt2=MPI_Wtime();

   error_chk();
   wdt=1.0e6*(wt2-wt1)/(double)(2*n*VOLUME);

   if (my_rank==0)
   {
      printf("Time per lattice point: %4.3f micro sec",wdt);
      printf(" (%d Mflops [%d bit arithmetic])\n\n",
             (int)(1896.0/wdt),(int)sizeof(spinor)/3);
      fclose(flog);
   }

   MPI_Finalize();
   exit(0);
}
