
/*******************************************************************************
*
* File check14.c
*
* Copyright (C) 2005, 2007 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Statistical test of the pseudo-fermion distributions
*
*******************************************************************************/

#define MAIN_PROGRAM

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "mpi.h"
#include "su3.h"
#include "random.h"
#include "start.h"
#include "misc.h"
#include "block.h"
#include "sw_term.h"
#include "dirac.h"
#include "hmc.h"
#include "dfl.h"
#include "global.h"

static lat_parms_t lat;
static gcr_parms_t gcr;
static hmc_parms_t hmc;
static dfl_parms_t dfl;


static double sum_action(hmc_act_t act)
{
   int nb,n;
   double *a0,*a1,*a2,sm,sall;

   nb=act.nb;
   a0=act.act0;
   a1=act.act1;
   a2=act.act2;   

   sm=0.0;

   for (n=0;n<nb;n++)
      sm+=(a0[n]+a1[n]+a2[n]);

   if (NPROC>1)
   {
      MPI_Reduce(&sm,&sall,1,MPI_DOUBLE,MPI_SUM,0,MPI_COMM_WORLD);
      MPI_Bcast(&sall,1,MPI_DOUBLE,0,MPI_COMM_WORLD);     
   }
   else
      sall=sm;

   return sall;
}


int main(int argc,char *argv[])
{
   int my_rank,nodfl,id;
   int bc,nc_first,nc_last,nc_step;   
   int bs_hmc[4],N0,N1,N2;
   int nmx_acf,nmx_gcr_acf,nmx_rop,nmx_frf,nmx_gcr_frf;
   int bs_sap[4],nmr_sap,ncy_sap,nkv_sap;
   int bs_dfl[4],Ns,rinv,rnmr,rncy,rvnmx;
   int vnkv,vnmx,vdnmx;
   int stat_rop,stat_acf,stat_gcr_acf[3],ifail;
   int stat_dfl_mds,dmy;
   int n,nb,isw,i0,v1,v2,*bs;
   int stat,flag,stat_min,stat_sum;
   int nsize,icnfg;
   double beta,kappa,csw;
   double tau,res_acf,res_gcr_acf,res_rop,res_frf,res_gcr_frf;
   double rkappa,rvres,vres,vdres;
   double blk_act,sact0,sact1;
   spinor_dble **wsd;
   char cnfg_dir[NAME_SIZE],cnfg_file[NAME_SIZE];
   char nbase[NAME_SIZE]; 
   hmc_act_t act0,act1;
   FILE *flog=NULL,*fin=NULL;

   MPI_Init(&argc,&argv);
   MPI_Comm_rank(MPI_COMM_WORLD,&my_rank);

   if (my_rank==0)
   {
      flog=freopen("check14.log","w",stdout);
      fin=freopen("check14.in","r",stdin);
      
      nodfl=find_opt(argc,argv,"-nodfl");      

      printf("\n");
      printf("Statistical test of the pseudo-fermion distributions\n");
      printf("----------------------------------------------------\n\n");

      printf("%dx%dx%dx%d lattice, ",NPROC0*L0,NPROC1*L1,NPROC2*L2,NPROC3*L3);
      printf("%dx%dx%dx%d process grid, ",NPROC0,NPROC1,NPROC2,NPROC3);
      printf("%dx%dx%dx%d local lattice\n\n",L0,L1,L2,L3);

      read_line("id","%d\n",&id);
      read_line("cnfg_dir","%s\n",cnfg_dir);       
      read_line("beta","%lf\n",&beta);
      read_line("kappa","%lf\n",&kappa);
      read_line("csw","%lf\n",&csw);
      read_line("bc","%d\n",&bc);
      read_line("nc_first","%d\n",&nc_first);
      read_line("nc_last","%d\n",&nc_last);
      read_line("nc_step","%d\n",&nc_step);
      
      read_line("bs_hmc","%d %d %d %d\n",
                &bs_hmc[0],&bs_hmc[1],&bs_hmc[2],&bs_hmc[3]);
      read_line("nmx_rop","%d\n",&nmx_rop);
      read_line("res_rop","%lf\n",&res_rop);
      read_line("nmx_acf","%d\n",&nmx_acf);
      read_line("res_acf","%lf\n",&res_acf);       
      read_line("nmx_gcr_acf","%d\n",&nmx_gcr_acf);
      read_line("res_gcr_acf","%lf\n",&res_gcr_acf);

      read_line("bs_sap","%d %d %d %d\n",
                &bs_sap[0],&bs_sap[1],&bs_sap[2],&bs_sap[3]);
      read_line("nmr_sap","%d\n",&nmr_sap);
      read_line("ncy_sap","%d\n",&ncy_sap);
      read_line("nkv_sap","%d\n",&nkv_sap);      

      if (nodfl==0)
      {
         read_line("bs_dfl","%d %d %d %d\n",
                   &bs_dfl[0],&bs_dfl[1],&bs_dfl[2],&bs_dfl[3]);
         read_line("Ns","%d\n",&Ns);
         read_line("rkappa","%lf\n",&rkappa);
         read_line("rinv","%d\n",&rinv);
         read_line("rnmr","%d\n",&rnmr);
         read_line("rncy","%d\n",&rncy);
         read_line("rvnmx","%d\n",&rvnmx);      
         read_line("rvres","%lf\n",&rvres);
         read_line("vnkv","%d\n",&vnkv);
         read_line("vnmx","%d\n",&vnmx);
         read_line("vres","%lf\n",&vres);
         read_line("vdnmx","%d\n",&vdnmx);
         read_line("vdres","%lf",&vdres);
      }

      fclose(fin);
   }

   MPI_Bcast(&nodfl,1,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&id,1,MPI_INT,0,MPI_COMM_WORLD);      
   MPI_Bcast(cnfg_dir,NAME_SIZE,MPI_CHAR,0,MPI_COMM_WORLD);
   MPI_Bcast(&beta,1,MPI_DOUBLE,0,MPI_COMM_WORLD);   
   MPI_Bcast(&kappa,1,MPI_DOUBLE,0,MPI_COMM_WORLD);
   MPI_Bcast(&csw,1,MPI_DOUBLE,0,MPI_COMM_WORLD);   
   MPI_Bcast(&bc,1,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&nc_first,1,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&nc_last,1,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&nc_step,1,MPI_INT,0,MPI_COMM_WORLD);
   
   MPI_Bcast(bs_hmc,4,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&nmx_rop,1,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&res_rop,1,MPI_DOUBLE,0,MPI_COMM_WORLD);
   MPI_Bcast(&nmx_acf,1,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&res_acf,1,MPI_DOUBLE,0,MPI_COMM_WORLD);
   MPI_Bcast(&nmx_gcr_acf,1,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&res_gcr_acf,1,MPI_DOUBLE,0,MPI_COMM_WORLD);
   MPI_Bcast(bs_sap,4,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&nmr_sap,1,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&ncy_sap,1,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&nkv_sap,1,MPI_INT,0,MPI_COMM_WORLD);   

   if (nodfl==0)
   {
      MPI_Bcast(bs_dfl,4,MPI_INT,0,MPI_COMM_WORLD);
      MPI_Bcast(&Ns,1,MPI_INT,0,MPI_COMM_WORLD);   
      MPI_Bcast(&rkappa,1,MPI_DOUBLE,0,MPI_COMM_WORLD);
      MPI_Bcast(&rinv,1,MPI_INT,0,MPI_COMM_WORLD);
      MPI_Bcast(&rnmr,1,MPI_INT,0,MPI_COMM_WORLD);   
      MPI_Bcast(&rncy,1,MPI_INT,0,MPI_COMM_WORLD);   
      MPI_Bcast(&rvnmx,1,MPI_INT,0,MPI_COMM_WORLD);   
      MPI_Bcast(&rvres,1,MPI_DOUBLE,0,MPI_COMM_WORLD);
      MPI_Bcast(&vnkv,1,MPI_INT,0,MPI_COMM_WORLD);   
      MPI_Bcast(&vnmx,1,MPI_INT,0,MPI_COMM_WORLD);   
      MPI_Bcast(&vres,1,MPI_DOUBLE,0,MPI_COMM_WORLD);
      MPI_Bcast(&vdnmx,1,MPI_INT,0,MPI_COMM_WORLD);   
      MPI_Bcast(&vdres,1,MPI_DOUBLE,0,MPI_COMM_WORLD);
   }   

   start_ranlux(0,123);
   geometry();
   alloc_ud();

   if (nodfl==0)
   {
      alloc_sd(6);
      
      if (Ns<(2*nkv_sap))
         alloc_s(2*nkv_sap+1);
      else
         alloc_s(Ns+2);

      alloc_u();
      alloc_sw();
      dfl=set_dfl_parms(bs_dfl,Ns,rinv,rnmr,rncy,rvnmx,vnkv,vnmx,vdnmx,
                        rkappa,rvres,vres,vdres,0,0);
   }
   else
   {
      alloc_sd(4);
      alloc_s(2*nkv_sap+1);
   }

   tau=0.5;
   N0=1;
   N1=1;
   N2=1;
   nmx_frf=1;
   nmx_gcr_frf=1;
   res_frf=1.0;
   res_gcr_frf=1.0;

   lat=set_lat_parms(beta,kappa,csw);
   gcr=set_gcr_parms(bs_sap,nmr_sap,ncy_sap,nkv_sap);
   hmc=set_hmc_parms(bs_hmc,bc,N0,N1,N2,nodfl==0,0,0,0,
                     nmx_acf,nmx_gcr_acf,nmx_rop,nmx_frf,nmx_gcr_frf,
                     tau,
                     res_acf,res_gcr_acf,res_rop,res_frf,res_gcr_frf);

   if (my_rank==0)
   {
      printf("beta = %.6f\n",lat.beta);   
      printf("kappa = %.6f\n",lat.kappa);      
      printf("csw = %.6f\n",lat.csw);
      printf("bc = %d\n\n",hmc.bc);
      
      printf("bs_hmc = %d %d %d %d\n",bs_hmc[0],bs_hmc[1],bs_hmc[2],bs_hmc[3]);
      printf("nmx_rop = %d\n",hmc.nmx_rop);            
      printf("res_rop = %.2e\n",hmc.res_rop);      
      printf("nmx_acf = %d\n",hmc.nmx_acf);            
      printf("res_acf = %.2e\n",hmc.res_acf);      
      printf("nmx_gcr_acf = %d\n",hmc.nmx_gcr_acf);            
      printf("res_gcr_acf = %.2e\n\n",hmc.res_gcr_acf);

      printf("bs_sap = %d %d %d %d\n",bs_sap[0],bs_sap[1],bs_sap[2],bs_sap[3]);
      printf("nmr_sap = %d\n",gcr.nmr);                  
      printf("ncy_sap = %d\n",gcr.ncy);
      printf("nkv_sap = %d\n\n",gcr.nkv);                        

      if (nodfl==0)
      {
         printf("bs_dfl = %d %d %d %d\n",
                bs_dfl[0],bs_dfl[1],bs_dfl[2],bs_dfl[3]);
         printf("Ns = %d\n",dfl.Ns);
         printf("rkappa = %.6f\n",dfl.rkappa);
         printf("rinv = %d\n",dfl.rinv);
         printf("rnmr = %d\n",dfl.rnmr);
         printf("rncy = %d\n",dfl.rncy);         
         printf("rvnmx = %d\n",dfl.rvnmx);
         printf("rvres = %.1e\n",dfl.rvres);
         printf("vnkv = %d\n",dfl.vnkv);         
         printf("vnmx = %d\n",dfl.vnmx);
         printf("vres = %.1e\n",dfl.vres);
         printf("vdnmx = %d\n",dfl.vdnmx);
         printf("vdres = %.1e\n\n",dfl.vdres);
      }

      printf("Configurations no %d -> %d in steps of %d\n\n",
             nc_first,nc_last,nc_step);
      fflush(flog);
   }

   alloc_hmc_action(&act0);
   alloc_hmc_action(&act1);

   error_root(((nc_last-nc_first)%nc_step)!=0,1,"main [check12.c]",
              "nc_last-nc_first is not a multiple of nc_step");
   check_dir_root(cnfg_dir);   

   nsize=name_size("%dx%dx%dx%db%.2fk%.5fc%.5fid%d",
                   NPROC0*L0,NPROC1*L1,NPROC2*L2,NPROC3*L3,beta,kappa,csw,id);
   error_root(nsize>=NAME_SIZE,1,"main [check12.c]","Base name is too long");
   sprintf(nbase,"%dx%dx%dx%db%.2fk%.5fc%.5fid%d",
           NPROC0*L0,NPROC1*L1,NPROC2*L2,NPROC3*L3,beta,kappa,csw,id);

   nsize=name_size("%s/%sn%d",cnfg_dir,nbase,nc_last);
   error_root(nsize>=NAME_SIZE,1,"main [check11.c]",
              "cnfg_dir name is too long");

   for (icnfg=nc_first;icnfg<=nc_last;icnfg+=nc_step)
   {
      if (my_rank==0)
      {
         printf("Configuration no %d:\n\n",icnfg);
         fflush(flog);
      }   

      sprintf(cnfg_file,"%s/%sn%d",cnfg_dir,nbase,icnfg);
      import_cnfg(cnfg_file);      

      start_md(act0,&stat_rop,&stat_dfl_mds,&ifail);
      sact0=sum_action(act0);

      if (my_rank==0)
      {
         printf("start_md:\n");
         if (nodfl)
            printf("stat_rop = %d, ifail = %d\n",stat_rop,ifail);
         else
            printf("stat_rop = %d, stat_dfl_mds = %d, ifail = %d\n",
                   stat_rop,stat_dfl_mds,ifail);
         printf("action = %.4e\n\n",sact0);
         fflush(flog);
      }

      error_root((stat_rop<0)||(stat_dfl_mds<0)||(ifail!=0),1,
                 "main [check14.c]","start_md() failed");

      blk_list(HMC_BLOCKS,&nb,&isw);

      blk_act=0.0;
      stat_acf=0;
      ifail=0;

      for (n=0;n<nb;n++)
      {
         blk_act+=hmc_acf(n,1,0,&stat,&flag);

         if (stat<0)
            stat_acf=stat;
         else
            stat_acf+=stat;

         ifail|=flag;
      
         if ((stat_acf<0)||(ifail!=0))
            break;
      }
                        
      MPI_Reduce(&stat_acf,&stat_min,1,MPI_INT,MPI_MIN,0,MPI_COMM_WORLD);
      MPI_Reduce(&stat_acf,&stat_sum,1,MPI_INT,MPI_SUM,0,MPI_COMM_WORLD);
      MPI_Reduce(&ifail,&flag,1,MPI_INT,MPI_MAX,0,MPI_COMM_WORLD);
      MPI_Bcast(&stat_min,1,MPI_INT,0,MPI_COMM_WORLD);      
      MPI_Bcast(&stat_sum,1,MPI_INT,0,MPI_COMM_WORLD);            
      MPI_Bcast(&flag,1,MPI_INT,0,MPI_COMM_WORLD);

      if (stat_min<0)
         stat_acf=stat_min;
      else
         stat_acf=stat_sum/(NPROC*nb);
   
      ifail=flag;

      error_root((stat_acf<0)||(ifail!=0),1,"main [check14.c]",
                 "Calculation of block action failed");

      MPI_Reduce(&blk_act,&sact1,1,MPI_DOUBLE,MPI_SUM,0,MPI_COMM_WORLD);
      MPI_Bcast(&sact1,1,MPI_DOUBLE,0,MPI_COMM_WORLD);

      n=6*NPROC*VOLUME;
   
      if (my_rank==0)
      {
         printf("Expectation value of the block pseudo-fermion action:\n");
         printf("stat_acf = %d, ifail = %d\n",stat_acf,ifail);
         printf("<action>/point = %.4f ",sact1/(double)(n));
         printf("(expect 1.0000+-%.4f)\n\n",sqrt(1.0/(double)(2*n)));
         fflush(flog);
      }

      wsd=reserve_wsd(1);
      i0=(wsd[0]-psd[0][0])/NSPIN;
      set_sd2zero(VOLUME,wsd[0]);
   
      hmc_gcr_acf(i0,0,act1.act2,&dmy,stat_gcr_acf,&ifail);
      sact1=sum_action(act1);

      release_wsd();

      error_root((stat_gcr_acf[0]<0)||(stat_gcr_acf[1]<0)||
                 (stat_gcr_acf[2]<0)||(ifail!=0),1,"main [check14.c]",
                 "Calculation of block-interaction action failed");
      error_chk();
   
      bs=bs_hmc;

      v2=2*((bs[0]-2)*(bs[1]-2)*(bs[2]-2)+
            (bs[0]-2)*(bs[1]-2)*(bs[3]-2)+
            (bs[0]-2)*(bs[2]-2)*(bs[3]-2)+
            (bs[1]-2)*(bs[2]-2)*(bs[3]-2));
   
      v1=bs[0]*bs[1]*bs[2]*bs[3]-(bs[0]-2)*(bs[1]-2)*(bs[2]-2)*(bs[3]-2)-v2;
      n=((12*v1+6*v2)*(NPROC*nb))/2;
   
      if (my_rank==0)
      {
         printf("Expectation value of the block-interaction action:\n");
         if (nodfl)
            printf("stat_gcr_acf = %d, ifail = %d\n",stat_gcr_acf[0],ifail);
         else
            printf("stat_gcr_acf = %d (little: %d, %d), ifail = %d\n",
                   stat_gcr_acf[0],stat_gcr_acf[1],stat_gcr_acf[2],ifail);
         printf("<action>/point = %.4f ",sact1/(double)(n));
         printf("(expect 1.0000+-%.4f)\n\n",sqrt(1.0/(double)(2*n)));
      }
   }

   if (my_rank==0)
      fclose(flog);   
   
   MPI_Finalize();    
   exit(0);
}
