
/*******************************************************************************
*
* File check3.c
*
* Copyright (C) 2005 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Check of hmc_acg() and hmc_frg()
*
*******************************************************************************/

#define MAIN_PROGRAM

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "mpi.h"
#include "su3.h"
#include "random.h"
#include "start.h"
#include "update.h"
#include "block.h"
#include "sap_gcr.h"
#include "hmc.h"
#include "global.h"

static const su3_dble ud0={{0.0}};


static void init_ud(void)
{
   su3_alg_dble X;
   su3_dble *u,*um;

   u=pud[VOLUME/2][0];
   um=u+4*VOLUME;

   for (;u<um;u++)
   {
      *u=ud0;
      (*u).c11.re=1.0;
      (*u).c22.re=1.0;
      (*u).c33.re=1.0;      
      random_alg(1,&X);
      ExpX(0.1,&X,u);      
   }

   set_flags(NEW_UD);
   renormalize_ud();
}


static void rot_ud(int n,double t)
{
   int nb,isw,*lk,*lkm;
   su3_dble *ub;
   su3_alg_dble *mom;
   block_t *b;
   hmc_t *hflds;

   b=blk_list(HMC_BLOCKS,&nb,&isw)+n;
   hflds=hmcflds()+n;
   
   ub=(*b).ud[0];
   mom=(*hflds).mom;
   lk=(*hflds).lks;
   lkm=lk+(*hflds).nlks;
   
   for (;lk<lkm;lk++)
   {
      ExpX(t,mom,ub+(*lk));
      mom+=1;
   }

   set_blk_flags(HMC_BLOCKS,NEW_ACTIVE_UD,n);
}


static double dSdt(int n)
{
   double sm;
   su3_alg_dble *fr,*frm,*mom;
   hmc_t *hflds;

   hflds=hmcflds()+n;

   fr=(*hflds).frc;
   frm=fr+(*hflds).nlks;
   mom=(*hflds).mom;   
   sm=0.0;
   
   for (;fr<frm;fr++)
   {
      sm+=12.0*((*mom).c1*(*fr).c1+(*mom).c2*(*fr).c2);
      sm-= 6.0*((*mom).c1*(*fr).c2+(*mom).c2*(*fr).c1);
      sm+= 4.0*((*mom).c3*(*fr).c3+(*mom).c4*(*fr).c4+(*mom).c5*(*fr).c5+
                (*mom).c6*(*fr).c6+(*mom).c7*(*fr).c7+(*mom).c8*(*fr).c8);
      mom+=1;
   }

   return sm;
}


int main(int argc,char *argv[])
{
   int my_rank,bs[4],nb,isw,n;
   double eps,actg,actb,act0,act1,dact,dsdt;
   double dev_frc,dev_frc_all,dev_frc_max;
   double sig_loss,sig_loss_all,sig_loss_max;   
   block_t *b;
   hmc_t *hflds;
   FILE *flog=NULL,*fin=NULL;

   MPI_Init(&argc,&argv);
   MPI_Comm_rank(MPI_COMM_WORLD,&my_rank);

   if (my_rank==0)
   {
      flog=freopen("check3.log","w",stdout);
      fin=freopen("check3.in","r",stdin);
      
      printf("\n");
      printf("Check of hmc_acg() and hmc_frg()\n");
      printf("--------------------------------\n\n");

      printf("%dx%dx%dx%d lattice, ",NPROC0*L0,NPROC1*L1,NPROC2*L2,NPROC3*L3);
      printf("%dx%dx%dx%d process grid, ",NPROC0,NPROC1,NPROC2,NPROC3);
      printf("%dx%dx%dx%d local lattice\n\n",L0,L1,L2,L3);

      scanf("bs %d %d %d %d",&bs[0],&bs[1],&bs[2],&bs[3]);      
      fclose(fin);

      printf("bs = %d %d %d %d\n",bs[0],bs[1],bs[2],bs[3]);      
      printf("Random gauge field\n\n");
   }

   MPI_Bcast(bs,4,MPI_INT,0,MPI_COMM_WORLD);
   
   start_ranlux(0,1234);
   geometry();
   alloc_ud();
   alloc_bgr(HMC_BLOCKS,bs);
   init_ud();

   assign_ud2udbgr(HMC_BLOCKS);
   actg=plaq_sum_dble();
   act0=0.0;
   act1=0.0;

   b=blk_list(HMC_BLOCKS,&nb,&isw);
   hflds=hmcflds();
   
   for (n=0;n<nb;n++)
   {
      act0+=hmc_acg(n);
      random_alg((*hflds).nlks,(*hflds).mom);
      rot_ud(n,0.1);
      act1+=hmc_acg(n);
   
      b+=1;      
      hflds+=1;
   }

   MPI_Reduce(&act0,&actb,1,MPI_DOUBLE,MPI_SUM,0,MPI_COMM_WORLD);
   act0=actb;
   MPI_Reduce(&act1,&actb,1,MPI_DOUBLE,MPI_SUM,0,MPI_COMM_WORLD);
   act1=actb;
   actb=act1-act0;
   sig_loss_max=-log10(fabs(1.0-act1/act0));

   act0=actg;
   swap_active(HMC_BLOCKS);
   act1=plaq_sum_dble();
   actg=act0-act1;
   sig_loss=-log10(fabs(1.0-act1/act0));
   if (sig_loss>sig_loss_max)
      sig_loss_max=sig_loss;
   actg*=2.0;
   
   if (my_rank==0)
   {
      printf("Calculation of the gauge action on the blocks:\n");
      printf("|1-sum_b[S(b)'-S(b)]/[S'-S]| = %.2e ",fabs(1.0-actb/actg));
      printf("[significance loss = %d digits]\n\n",(int)(sig_loss_max));
   }

   b=blk_list(HMC_BLOCKS,&nb,&isw);
   hflds=hmcflds();   
   dev_frc_max=0.0;
   sig_loss_max=0.0;
   eps=1.0e-4;
   
   for (n=0;n<nb;n++)
   {
      hmc_frg(n);
      dsdt=dSdt(n);

      rot_ud(n,eps);
      act0=2.0*hmc_acg(n)/3.0;
      rot_ud(n,-eps);

      rot_ud(n,-eps);
      act1=2.0*hmc_acg(n)/3.0;
      rot_ud(n,eps);
      
      rot_ud(n,2.0*eps);
      act0-=hmc_acg(n)/12.0;
      rot_ud(n,-2.0*eps);

      rot_ud(n,-2.0*eps);
      act1-=hmc_acg(n)/12.0;
      rot_ud(n,2.0*eps);

      dact=(act0-act1)/eps;
      dev_frc=fabs(1.0-dsdt/dact);

      MPI_Reduce(&dev_frc,&dev_frc_all,1,MPI_DOUBLE,MPI_MAX,0,MPI_COMM_WORLD);
      if (dev_frc_all>dev_frc_max)
         dev_frc_max=dev_frc_all;

      sig_loss=-log10(fabs(1.0-act0/act1));
      MPI_Reduce(&sig_loss,&sig_loss_all,1,MPI_DOUBLE,MPI_MAX,0,MPI_COMM_WORLD);
      if (sig_loss_all>sig_loss_max)
         sig_loss_max=sig_loss_all;      

      b+=1;
      hflds+=1;
   }
   
   error_chk();

   if (my_rank==0)
   {
      printf("Calculation of the force:\n");
      printf("Relative deviation of dS/dt = %.2e ",dev_frc_max);
      printf("[significance loss = %d digits]\n\n",(int)(sig_loss_max));
      fclose(flog);
   }

   MPI_Finalize();    
   exit(0);
}
