
/*******************************************************************************
*
* File check9.c
*
* Copyright (C) 2005, 2007 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Check of ROp()
*
*******************************************************************************/

#define MAIN_PROGRAM

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "mpi.h"
#include "su3.h"
#include "random.h"
#include "start.h"
#include "misc.h"
#include "block.h"
#include "linalg.h"
#include "dirac.h"
#include "sw_term.h"
#include "sap_gcr.h"
#include "hmc.h"
#include "dfl.h"
#include "global.h"

static const su3_dble ud0={{0.0}};
static lat_parms_t lat;
static hmc_parms_t hmc;
static gcr_parms_t gcr;
static dfl_parms_t dfl;


int main(int argc,char *argv[])
{
   int my_rank,nodfl,id;
   int bc,nc_first,nc_last,nc_step;
   int bs_hmc[4],N0,N1,N2;
   int nmx_acf,nmx_gcr_acf,nmx_rop,nmx_frf,nmx_gcr_frf;
   int bs_sap[4],nmr_sap,ncy_sap,nkv_sap;
   int bs_dfl[4],Ns,rinv,rnmr,rncy,rvnmx;
   int vnkv,vnmx,vdnmx;
   int nb,isw,nsize,icnfg,status[3],ifail;
   int i0,i1,i2,i3,i4;
   double beta,kappa,csw;
   double tau,res_acf,res_gcr_acf,res_rop,res_frf,res_gcr_frf;
   double rkappa,rvres,vres,vdres;
   double dmax,res,cnrm;
   complex_dble z;
   spinor_dble **wsd,**rsd;
   block_t *b,*bm;
   char cnfg_dir[NAME_SIZE],cnfg_file[NAME_SIZE];
   char nbase[NAME_SIZE];
   FILE *flog=NULL,*fin=NULL;

   MPI_Init(&argc,&argv);
   MPI_Comm_rank(MPI_COMM_WORLD,&my_rank);

   if (my_rank==0)
   {
      flog=freopen("check9.log","w",stdout);
      fin=freopen("check9.in","r",stdin);

      nodfl=find_opt(argc,argv,"-nodfl");      
      
      printf("\n");
      printf("Check of the program ROp\n");
      printf("------------------------\n\n");

      printf("%dx%dx%dx%d lattice, ",NPROC0*L0,NPROC1*L1,NPROC2*L2,NPROC3*L3);
      printf("%dx%dx%dx%d process grid, ",NPROC0,NPROC1,NPROC2,NPROC3);
      printf("%dx%dx%dx%d local lattice\n\n",L0,L1,L2,L3);

      read_line("id","%d\n",&id);
      read_line("cnfg_dir","%s\n",cnfg_dir);
      read_line("beta","%lf\n",&beta);      
      read_line("kappa","%lf\n",&kappa);
      read_line("csw","%lf\n",&csw);
      read_line("bc","%d\n",&bc);
      read_line("nc_first","%d\n",&nc_first);
      read_line("nc_last","%d\n",&nc_last);    
      read_line("nc_step","%d\n",&nc_step);    
      
      read_line("bs_hmc","%d %d %d %d\n",
                &bs_hmc[0],&bs_hmc[1],&bs_hmc[2],&bs_hmc[3]);
      read_line("nmx_rop","%d\n",&nmx_rop);
      read_line("res_rop","%lf\n",&res_rop);
      read_line("nmx_gcr_acf","%d\n",&nmx_gcr_acf);
      read_line("res_gcr_acf","%lf\n",&res_gcr_acf);

      read_line("bs_sap","%d %d %d %d\n",
                &bs_sap[0],&bs_sap[1],&bs_sap[2],&bs_sap[3]);
      read_line("nmr_sap","%d\n",&nmr_sap);
      read_line("ncy_sap","%d\n",&ncy_sap);
      read_line("nkv_sap","%d\n",&nkv_sap);

      if (nodfl==0)
      {
         read_line("bs_dfl","%d %d %d %d\n",
                   &bs_dfl[0],&bs_dfl[1],&bs_dfl[2],&bs_dfl[3]);
         read_line("Ns","%d\n",&Ns);
         read_line("rkappa","%lf\n",&rkappa);
         read_line("rinv","%d\n",&rinv);
         read_line("rnmr","%d\n",&rnmr);
         read_line("rncy","%d\n",&rncy);
         read_line("rvnmx","%d\n",&rvnmx);      
         read_line("rvres","%lf\n",&rvres);
         read_line("vnkv","%d\n",&vnkv);
         read_line("vnmx","%d\n",&vnmx);
         read_line("vres","%lf\n",&vres);
         read_line("vdnmx","%d\n",&vdnmx);
         read_line("vdres","%lf",&vdres);
      }
      
      fclose(fin);
   }

   MPI_Bcast(&nodfl,1,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&id,1,MPI_INT,0,MPI_COMM_WORLD);      
   MPI_Bcast(cnfg_dir,NAME_SIZE,MPI_CHAR,0,MPI_COMM_WORLD);
   MPI_Bcast(&beta,1,MPI_DOUBLE,0,MPI_COMM_WORLD);   
   MPI_Bcast(&kappa,1,MPI_DOUBLE,0,MPI_COMM_WORLD);
   MPI_Bcast(&csw,1,MPI_DOUBLE,0,MPI_COMM_WORLD);   
   MPI_Bcast(&bc,1,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&nc_first,1,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&nc_last,1,MPI_INT,0,MPI_COMM_WORLD); 
   MPI_Bcast(&nc_step,1,MPI_INT,0,MPI_COMM_WORLD); 

   MPI_Bcast(bs_hmc,4,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&nmx_rop,1,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&res_rop,1,MPI_DOUBLE,0,MPI_COMM_WORLD);
   MPI_Bcast(&nmx_gcr_acf,1,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&res_gcr_acf,1,MPI_DOUBLE,0,MPI_COMM_WORLD);   
   MPI_Bcast(bs_sap,4,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&nmr_sap,1,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&ncy_sap,1,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&nkv_sap,1,MPI_INT,0,MPI_COMM_WORLD);   

   if (nodfl==0)
   {
      MPI_Bcast(bs_dfl,4,MPI_INT,0,MPI_COMM_WORLD);
      MPI_Bcast(&Ns,1,MPI_INT,0,MPI_COMM_WORLD);   
      MPI_Bcast(&rkappa,1,MPI_DOUBLE,0,MPI_COMM_WORLD);
      MPI_Bcast(&rinv,1,MPI_INT,0,MPI_COMM_WORLD);
      MPI_Bcast(&rnmr,1,MPI_INT,0,MPI_COMM_WORLD);   
      MPI_Bcast(&rncy,1,MPI_INT,0,MPI_COMM_WORLD);   
      MPI_Bcast(&rvnmx,1,MPI_INT,0,MPI_COMM_WORLD);   
      MPI_Bcast(&rvres,1,MPI_DOUBLE,0,MPI_COMM_WORLD);
      MPI_Bcast(&vnkv,1,MPI_INT,0,MPI_COMM_WORLD);   
      MPI_Bcast(&vnmx,1,MPI_INT,0,MPI_COMM_WORLD);   
      MPI_Bcast(&vres,1,MPI_DOUBLE,0,MPI_COMM_WORLD);
      MPI_Bcast(&vdnmx,1,MPI_INT,0,MPI_COMM_WORLD);   
      MPI_Bcast(&vdres,1,MPI_DOUBLE,0,MPI_COMM_WORLD);
   }  
   
   start_ranlux(0,1234);
   geometry();
   alloc_ud();

   if (nodfl==0)
   {
      alloc_sd(8);
      
      if (Ns<(2*nkv_sap))
         alloc_s(2*nkv_sap+1);
      else
         alloc_s(Ns+2);

      alloc_u();
      alloc_sw();
      dfl=set_dfl_parms(bs_dfl,Ns,rinv,rnmr,rncy,rvnmx,vnkv,vnmx,vdnmx,
                        rkappa,rvres,vres,vdres,0,0);
   }
   else
   {
      alloc_sd(6);
      alloc_s(2*nkv_sap+1);
   }

   N0=1;
   N1=1;
   N2=1;
   nmx_acf=1;
   nmx_frf=1;
   nmx_gcr_frf=1;
   tau=0.0;
   res_acf=1.0;
   res_frf=1.0;
   res_gcr_frf=1.0;

   lat=set_lat_parms(beta,kappa,csw);
   gcr=set_gcr_parms(bs_sap,nmr_sap,ncy_sap,nkv_sap);
   hmc=set_hmc_parms(bs_hmc,bc,N0,N1,N2,nodfl==0,0,0,0,
                     nmx_acf,nmx_gcr_acf,nmx_rop,nmx_frf,nmx_gcr_frf,
                     tau,
                     res_acf,res_gcr_acf,res_rop,res_frf,res_gcr_frf);

   if (my_rank==0)
   {
      printf("beta = %.6f\n",lat.beta);
      printf("kappa = %.6f\n",lat.kappa);      
      printf("csw = %.6f\n",lat.csw);
      printf("bc = %d\n\n",hmc.bc);     
      
      printf("bs_hmc = %d %d %d %d\n",bs_hmc[0],bs_hmc[1],bs_hmc[2],bs_hmc[3]);
      printf("nmx_rop = %d\n",hmc.nmx_rop);            
      printf("res_rop = %.2e\n",hmc.res_rop);      
      printf("nmx_gcr_acf = %d\n",hmc.nmx_gcr_acf);            
      printf("res_gcr_acf = %.2e\n\n",hmc.res_gcr_acf);

      printf("bs_sap = %d %d %d %d\n",bs_sap[0],bs_sap[1],bs_sap[2],bs_sap[3]);
      printf("nmr_sap = %d\n",gcr.nmr);                  
      printf("ncy_sap = %d\n",gcr.ncy);
      printf("nkv_sap = %d\n\n",gcr.nkv);                        

      if (nodfl==0)
      {
         printf("bs_dfl = %d %d %d %d\n",
                bs_dfl[0],bs_dfl[1],bs_dfl[2],bs_dfl[3]);
         printf("Ns = %d\n",dfl.Ns);
         printf("rkappa = %.6f\n",dfl.rkappa);
         printf("rinv = %d\n",dfl.rinv);
         printf("rnmr = %d\n",dfl.rnmr);
         printf("rncy = %d\n",dfl.rncy);         
         printf("rvnmx = %d\n",dfl.rvnmx);
         printf("rvres = %.1e\n",dfl.rvres);
         printf("vnkv = %d\n",dfl.vnkv);         
         printf("vnmx = %d\n",dfl.vnmx);
         printf("vres = %.1e\n",dfl.vres);
         printf("vdnmx = %d\n",dfl.vdnmx);
         printf("vdres = %.1e\n\n",dfl.vdres);
      }

      printf("Configurations no %d -> %d in steps of %d\n\n",
             nc_first,nc_last,nc_step);
      fflush(flog);
   }

   error_root(((nc_last-nc_first)%nc_step)!=0,1,"main [check12.c]",
              "nc_last-nc_first is not a multiple of nc_step");
   check_dir_root(cnfg_dir);   

   nsize=name_size("%dx%dx%dx%db%.2fk%.5fc%.5fid%d",
                   NPROC0*L0,NPROC1*L1,NPROC2*L2,NPROC3*L3,beta,kappa,csw,id);
   error_root(nsize>=NAME_SIZE,1,"main [check12.c]","Base name is too long");
   sprintf(nbase,"%dx%dx%dx%db%.2fk%.5fc%.5fid%d",
           NPROC0*L0,NPROC1*L1,NPROC2*L2,NPROC3*L3,beta,kappa,csw,id);

   nsize=name_size("%s/%sn%d",cnfg_dir,nbase,nc_last);
   error_root(nsize>=NAME_SIZE,1,"main [check10.c]",
              "cnfg_dir name is too long");

   for (icnfg=nc_first;icnfg<=nc_last;icnfg+=nc_step)
   {
      if (my_rank==0)
      {
         printf("Configuration no %d:\n\n",icnfg);
         fflush(flog);
      }

      sprintf(cnfg_file,"%s/%sn%d",cnfg_dir,nbase,icnfg);
      import_cnfg(cnfg_file);   
   
      if (bc==-1)
         flipbcd();   
      assign_ud2udbgr(HMC_BLOCKS);
      swb_term();

      if (nodfl==0)
      {
         dfl_modes(0,status,&ifail);

         error_root((status[0]<0)||(ifail!=0),1,"main [check9.c]",
                    "Deflation subspace generation failed");
      }
      
      wsd=reserve_wsd(4);
      i0=(wsd[0]-psd[0][0])/NSPIN;
      i1=(wsd[1]-psd[0][0])/NSPIN;
      i2=(wsd[2]-psd[0][0])/NSPIN;
      i3=(wsd[3]-psd[0][0])/NSPIN;
   
      random_chi();
      assign_chi2sd(i0);
      cnrm=norm_square_dble(VOLUME,1,wsd[0]);
      ROp(status,&ifail);

      if (my_rank==0)
         printf("Application of ROp: status = %d, ifail = %d\n\n",
                status[0],ifail);

      assign_chi2sd(i1);
      set_sd2zero(VOLUME,wsd[2]);
      b=blk_list(HMC_BLOCKS,&nb,&isw);
      bm=b+nb;

      for (;b<bm;b++)
      {
         assign_sd2sdb(b,ALL_PTS,i1,0);
         Qbnds_dble(b,0,0);
         sub_assign_sdbb2sd(b,0,i2);
      } 

      mulmg5_dble(VOLUME,wsd[2]);

      if (nodfl)
      {
         res=sap_gcr(hmc.nmx_gcr_acf,hmc.res_gcr_acf,i2,i3,status,&ifail);
         status[1]=0;
         status[2]=0;
      }
      else
         res=dfl_sap_gcr(hmc.nmx_gcr_acf,hmc.res_gcr_acf,i2,i3,status,&ifail);

      if (my_rank==0)
      {
         printf("Solution of the Dirac equation:\n");

         if (nodfl)
            printf("status = %d, ifail = %d\n",status[0],ifail);
         else
            printf("status = %d (little: %d, %d), ifail = %d\n",
                   status[0],status[1],status[2],ifail);
      }
      
      error_root((status[0]<0)||(status[1]<0)||(status[2]<0)||(ifail!=0),1,
                 "main [check9.c]","SAP+GCR solver failed");
      
      rsd=reserve_wsd(1);
      i4=(rsd[0]-psd[0][0])/NSPIN;   
   
      Qnohat_dble(i3,i4);
      mulg5_dble(VOLUME,rsd[0]);
      z.re=-1.0;
      z.im=0.0;
      mulc_spinor_add_dble(VOLUME,rsd[0],wsd[2],z);
      dmax=norm_square_dble(VOLUME,1,rsd[0])/
         norm_square_dble(VOLUME,1,wsd[2]);   
      
      if (my_rank==0)
         printf("Residue = %.1e (should be <=%.1e)\n\n",
                sqrt(dmax),hmc.res_gcr_acf);

      sub_assign_PdOs(i1,i3);
      mulc_spinor_add_dble(VOLUME,wsd[0],wsd[1],z);
      dmax=norm_square_dble(VOLUME,1,wsd[0])/cnrm;
      error_chk();
   
      if (my_rank==0)
      {
         printf("||chi-R^(-1)*ROp*chi||/||chi|| = %.1e\n\n",sqrt(dmax));
         fflush(flog);
      }

      release_wsd();
      release_wsd();
   }

   if (my_rank==0)
      fclose(flog);   
   
   MPI_Finalize();   
   exit(0);
}
