
/*******************************************************************************
*
* File time4.c
*
* Copyright (C) 2007 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Timing of the valg_dbl routines
*
*******************************************************************************/

#define MAIN_PROGRAM

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <time.h>
#include "mpi.h"
#include "random.h"
#include "start.h"
#include "misc.h"
#include "linalg.h"
#include "dfl.h"
#include "global.h"

static complex_dble v[25];
static complex_dble w[25];
static complex_dble *ppk[10];


int main(int argc,char *argv[])
{
   int my_rank,count,i,j,icom;
   int bs[4],nm,n,nv,nvec;
   double r,wt1,wt2,wdt;
   complex_dble z,mz;
   complex_dble **wv,*pk,*pl;
   FILE *flog=NULL,*fin=NULL;   

   MPI_Init(&argc,&argv);
   MPI_Comm_rank(MPI_COMM_WORLD,&my_rank);

   if (my_rank==0)
   {
      flog=freopen("time4.log","w",stdout);
      fin=freopen("time4.in","r",stdin);
      
      printf("\n");
      printf("Timing of the valg_dble routines\n");
      printf("--------------------------------\n\n");   
      
      printf("%dx%dx%dx%d lattice, ",NPROC0*L0,NPROC1*L1,NPROC2*L2,NPROC3*L3);
      printf("%dx%dx%dx%d process grid, ",NPROC0,NPROC1,NPROC2,NPROC3);
      printf("%dx%dx%dx%d local lattice\n\n",L0,L1,L2,L3);

      read_line("bs","%d %d %d %d",&bs[0],&bs[1],&bs[2],&bs[3]);
      read_line("nm","%d",&nm);
      fclose(fin);

      printf("bs = %d %d %d %d\n",bs[0],bs[1],bs[2],bs[3]);
      printf("nm = %d\n\n",nm);
      fflush(flog);      
   }

   MPI_Bcast(bs,4,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&nm,1,MPI_INT,0,MPI_COMM_WORLD);   

   start_ranlux(0,12345);
   geometry();
   set_dfl_parms(bs,nm,4,1,1,1,1,1,1,0.1,1.0e-3,1.0e-6,1.0e-12,0,10);
   wv=reserve_wvd(10);
   vdflds_base(&n,&nv,&nvec);
   icom=1;

   for (i=0;i<10;i++)
      random_vd(nv,wv[i],1.0f);

   n=(int)(6.0e7/(double)(nv));   
   if (n<1)
      n=1;

   pk=wv[0];
   pl=wv[1];

   MPI_Barrier(MPI_COMM_WORLD);
   wt1=MPI_Wtime(); 
   for (count=0;count<n;count++)
      z=vprod_dble(nv,icom,pk,pl);
   MPI_Barrier(MPI_COMM_WORLD);
   wt2=MPI_Wtime();   

   wdt=1.0e6*(wt2-wt1)/(double)(n*nv);

   if (my_rank==0)
   {   
      printf("Function vprod_dble:\n");
      printf("Time per block: %4.3f micro sec\n",(double)(nm)*wdt);
      printf("%d Mflops [%d bit arithmetic]\n\n",
             (int)(8.0/wdt),4*sizeof(complex_dble));
   }
   
   MPI_Barrier(MPI_COMM_WORLD);
   wt1=MPI_Wtime();   
   for (count=0;count<n;count++)
      r=vnorm_square_dble(nv,icom,pk);
   MPI_Barrier(MPI_COMM_WORLD);
   wt2=MPI_Wtime();   

   wdt=1.0e6*(wt2-wt1)/(double)(n*nv);

   if (my_rank==0)
   {   
      printf("Function vnorm_square_dble:\n");
      printf("Time per block: %4.3f micro sec\n",(double)(nm)*wdt);
      printf("%d Mflops [%d bit arithmetic]\n\n",
             (int)(4.0/wdt),4*sizeof(complex_dble));
   }
   
   MPI_Barrier(MPI_COMM_WORLD);
   wt1=MPI_Wtime();   
   for (count=0;count<n;count++)
      vnormalize_dble(nv,icom,pk);
   MPI_Barrier(MPI_COMM_WORLD);
   wt2=MPI_Wtime();   

   wdt=1.0e6*(wt2-wt1)/(double)(n*nv);

   if (my_rank==0)
   {   
      printf("Function vnormalize_dble:\n");
      printf("Time per block: %4.3f micro sec\n",(double)(nm)*wdt);
      printf("%d Mflops [%d bit arithmetic]\n\n",
             (int)(6.0/wdt),4*sizeof(complex_dble));
   }
   
   mz.re=-z.re;
   mz.im=-z.im;

   MPI_Barrier(MPI_COMM_WORLD);
   wt1=MPI_Wtime();    
   for (count=0;count<n;count++)
   {
      mulc_vadd_dble(nv,pk,pl,z);
      mulc_vadd_dble(nv,pk,pl,mz);
   }
   MPI_Barrier(MPI_COMM_WORLD);
   wt2=MPI_Wtime();   

   wdt=1.0e6*(wt2-wt1)/(double)(2*n*nv);

   if (my_rank==0)
   {   
      printf("Function mulc_vadd_dble:\n");
      printf("Time per block: %4.3f micro sec\n",(double)(nm)*wdt);
      printf("%d Mflops [%d bit arithmetic]\n\n",
             (int)(8.0/wdt),4*sizeof(complex_dble));
   }
   
   random_vd(nv,pk,1.0f);
   random_vd(nv,pl,1.0f);
   vnormalize_dble(nv,icom,pk);
   vnormalize_dble(nv,icom,pl);
   vproject_dble(nv,icom,pl,pk);
   
   MPI_Barrier(MPI_COMM_WORLD);
   wt1=MPI_Wtime();    
   for (count=0;count<n;count++)
      vproject_dble(nv,icom,pk,pl);
   MPI_Barrier(MPI_COMM_WORLD);
   wt2=MPI_Wtime();   

   wdt=1.0e6*(wt2-wt1)/(double)(n*nv);

   if (my_rank==0)
   {   
      printf("Function vproject_dble:\n");
      printf("Time per block: %4.3f micro sec\n",(double)(nm)*wdt);
      printf("%d Mflops [%d bit arithmetic]\n\n",
             (int)(16.0/wdt),4*sizeof(complex_dble));
   }
   
   for (i=0;i<5;i++)
   {
      random_vd(nv,wv[i],1.0f);
      ppk[i]=wv[i];

      for (j=0;j<i;j++)
         vproject_dble(nv,icom,ppk[i],ppk[j]);

      vnormalize_dble(nv,icom,ppk[i]);
   }

   for (i=5;i<10;i++)
   {
      random_vd(nv,wv[i],1.0f);
      ppk[i]=wv[i];

      for (j=5;j<i;j++)
         vproject_dble(nv,icom,ppk[i],ppk[j]);

      vnormalize_dble(nv,icom,ppk[i]);
   }   
      
   for (i=0;i<5;i++)
   {
      for (j=0;j<5;j++)
      {
         z=vprod_dble(nv,icom,ppk[i],ppk[j+5]);

         v[5*i+j].re= z.re;
         v[5*i+j].im= z.im;
         w[5*j+i].re= z.re;
         w[5*j+i].im=-z.im;         
      }
   }

   n/=25;
   if (n<1)
      n=1;
   
   MPI_Barrier(MPI_COMM_WORLD);
   wt1=MPI_Wtime();    
   for (count=0;count<n;count++)
   {
      if ((count&0x7)==0x0)
      {
         for (i=0;i<5;i++)
            assign_vd2vd(nv,ppk[i+5],ppk[i]);
      }
      
      vrotate_dble(nv,5,ppk,v);
      vrotate_dble(nv,5,ppk,w);
   }
   MPI_Barrier(MPI_COMM_WORLD);
   wt2=MPI_Wtime();   

   error_chk();   
   wdt=1.0e6*(wt2-wt1)/(double)(50*n*nv);
   
   if (my_rank==0)
   {   
      printf("Function vrotate_dble (n=5 fields):\n");
      printf("Time per block: %4.3f*n^2 micro sec\n",
             (double)(nm)*wdt);
      printf("%d Mflops [%d bit arithmetic]\n\n",
             (int)(8.0/wdt),4*sizeof(complex_dble));
      fclose(flog);
   }

   MPI_Finalize();   
   exit(0);
}
