
/*******************************************************************************
*
* File time4.c
*
* Copyright (C) 2005 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Timing of su3xsu3, su3dagxsu3, ...
*
*******************************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <time.h>
#include "random.h"
#include "su3.h"
#include "start.h"
#include "misc.h"

#define N 10000000


int main(void)
{
   int count;
   double t1,t2,dt;
   su3_dble *u,*v,*w;
   u3_alg_dble *X;

   printf("\n");
   printf("Timing of su3xsu3, su3dagxsu3, ...\n");
   printf("----------------------------------\n\n");
   
   u=amalloc(3*sizeof(su3_dble),4);
   X=amalloc(sizeof(u3_alg_dble),3);
   error((u==NULL)||(X==NULL),1,"main [time4.c]",
         "Unable to allocate auxiliary array");
   v=u+1;
   w=u+2;

   rlxd_init(1,23456);   
   random_su3_dble(u);
   random_su3_dble(v);
   ranlxd((double*)(&(*X).c1),9);
   printf("The times per application are:\n");
   
   t1=(double)clock();
   for (count=0;count<N;count++)
      su3xsu3(u,v,w);
   t2=(double)clock();

   dt=(t2-t1)/(double)(CLOCKS_PER_SEC);
   dt*=1.0e6/(double)(N);

   printf("su3xsu3:       %4.3f micro sec (%d Mflops)\n",dt,(int)(198.0/dt));

   t1=(double)clock();
   for (count=0;count<N;count++)
      su3dagxsu3(u,v,w);
   t2=(double)clock();

   dt=(t2-t1)/(double)(CLOCKS_PER_SEC);
   dt*=1.0e6/(double)(N);

   printf("su3dagxsu3:    %4.3f micro sec (%d Mflops)\n",dt,(int)(198.0/dt));

   t1=(double)clock();
   for (count=0;count<N;count++)
      su3xsu3dag(u,v,w);
   t2=(double)clock();

   dt=(t2-t1)/(double)(CLOCKS_PER_SEC);
   dt*=1.0e6/(double)(N);

   printf("su3xsu3dag:    %4.3f micro sec (%d Mflops)\n",dt,(int)(198.0/dt));

   t1=(double)clock();
   for (count=0;count<N;count++)
      su3dagxsu3dag(u,v,w);
   t2=(double)clock();

   dt=(t2-t1)/(double)(CLOCKS_PER_SEC);
   dt*=1.0e6/(double)(N);

   printf("su3dagxsu3dag: %4.3f micro sec (%d Mflops)\n",dt,(int)(198.0/dt));

   t1=(double)clock();
   for (count=0;count<N;count++)
      su3xu3alg(u,X,v);
   t2=(double)clock();

   dt=(t2-t1)/(double)(CLOCKS_PER_SEC);
   dt*=1.0e6/(double)(N);

   printf("su3xu3alg:     %4.3f micro sec (%d Mflops)\n",dt,(int)(198.0/dt));

   t1=(double)clock();
   for (count=0;count<N;count++)
      su3dagxu3alg(u,X,v);
   t2=(double)clock();

   dt=(t2-t1)/(double)(CLOCKS_PER_SEC);
   dt*=1.0e6/(double)(N);

   printf("su3dagxu3alg:  %4.3f micro sec (%d Mflops)\n",dt,(int)(198.0/dt));

   t1=(double)clock();
   for (count=0;count<N;count++)
      u3algxsu3(X,u,v);
   t2=(double)clock();

   dt=(t2-t1)/(double)(CLOCKS_PER_SEC);
   dt*=1.0e6/(double)(N);

   printf("u3algxsu3:     %4.3f micro sec (%d Mflops)\n",dt,(int)(198.0/dt));

   t1=(double)clock();
   for (count=0;count<N;count++)
      u3algxsu3dag(X,u,v);
   t2=(double)clock();

   dt=(t2-t1)/(double)(CLOCKS_PER_SEC);
   dt*=1.0e6/(double)(N);

   printf("u3algxsu3dag:  %4.3f micro sec (%d Mflops)\n\n",dt,(int)(198.0/dt));
   exit(0);
}
