
/*******************************************************************************
*
* File time7.c
*
* Copyright (C) 2007 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Timing of cmat_vec_dble, cmat_mul_dble and cmat_inv_dble
*
*******************************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <time.h>
#include "random.h"
#include "su3.h"
#include "start.h"
#include "misc.h"


static void rvec(int n,complex_dble *v)
{
   gauss_dble((double*)(v),2*n);
}


static void rmat(int n,complex_dble *a)
{
   int i,j;
   double r;

   r=0.1/(double)(n*n);

   gauss_dble((double*)(a),2*n*n);

   for (i=0;i<n;i++)
   {
      for (j=0;j<n;j++)
      {
         a[i*n+j].re*=r;
         a[i*n+j].im*=r;
      }

      a[i*n+i].re+=1.0;
   }
}


int main(void)
{
   int n,nm,nsq,nr,ie;
   int i,count;
   double t1,t2,dt,k;
   complex_dble *a,*b,*c,*v,*w;

   printf("\n");
   printf("Timing of cmat_vec_dble, cmat_mul_dble and cmat_inv_dble\n");
   printf("--------------------------------------------------------\n\n");

   printf("Matrix size: ");
   scanf(" %d",&n);

   error(n<1,1,"main [time7.c]","Improper matrix size");

   nm=(int)(1.0e7/(double)(n*n*sizeof(complex_dble)))+1;
   a=amalloc(nm*n*(3*n+2)*sizeof(*a),4);
   error(a==NULL,1,"main [time7.c]","Unable to allocate auxiliary arrays");

   nsq=n*n;
   b=a+nm*nsq;
   c=b+nm*nsq;
   v=c+nm*nsq;
   w=v+nm*n;
   
   for (i=0;i<nm;i++)
   {
      rmat(n,a+i*nsq);
      rmat(n,b+i*nsq);
      rmat(n,c+i*nsq);
      rvec(n,v+i*n);
      rvec(n,w+i*n);      
   }

   nr=(int)(1.0e9/(double)(nm*nsq*4)+1);

   printf("\n");
   printf("The times per application are:\n\n");
   
   t1=(double)clock();
   for (count=0;count<nr;count++)
   {
      for (i=0;i<nm;i++)
         cmat_vec_dble(n,a+i*nsq,v+i*n,w+i*n);
   }
   t2=(double)clock();

   dt=(t2-t1)/(double)(CLOCKS_PER_SEC);
   dt*=1.0e6/(double)(nr*nm);

   printf("cmat_vec_dble:   %4.3f micro sec (%d Mflops)\n",
          dt,(int)((double)(n*((n-1)*8+6))/dt));

   nr=(int)(1.0e9/(double)(nm*nsq*n*4)+1);
   
   t1=(double)clock();
   for (count=0;count<nr;count++)
   {
      for (i=0;i<nm;i++)
         cmat_mul_dble(n,a+i*nsq,b+i*nsq,c+i*nsq);
   }
   t2=(double)clock();

   dt=(t2-t1)/(double)(CLOCKS_PER_SEC);
   dt*=1.0e6/(double)(nr*nm);

   printf("cmat_mul_dble:   %4.3f micro sec (%d Mflops)\n",
          dt,(int)((double)(n*n*((n-1)*8+6))/dt));   

   ie=0;
   
   t1=(double)clock();
   for (count=0;count<nr;count++)
   {
      for (i=0;i<nm;i++)
         ie+=cmat_inv_dble(n,a+i*nsq,b+i*nsq,&k);
   }
   t2=(double)clock();

   dt=(t2-t1)/(double)(CLOCKS_PER_SEC);
   dt*=1.0e6/(double)(nr*nm);

   error(ie!=0,1,"main [time7.c]","Some matrix inversions failed");
   
   printf("cmat_inv_dble:   %4.3f micro sec (~%d Mflops)\n\n",
          dt,(int)((double)(12*n*n*n)/dt));   
   
   exit(0);
}
