
/*******************************************************************************
*
* File check5.c
*
* Copyright (C) 2005 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Kolmogorov-Smirnov test of the random distribution produced by the
* function random_su2
*
*******************************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "random.h"
#include "extras.h"
#include "start.h"

static double rho;


static double prob1(double x)
{
   if ((x<=-1.0)||(x>=1.0))
      return(0.0);
   else
      return(sqrt(1.0-x*x)*exp(rho*(x-1.0)));
}


static double prob2(double x)
{
   if ((x<=-1.0)||(x>=1.0))
      return(0.0);
   else
   {
      if (rho<1.0e-8)
         return(sqrt(1.0-x*x));
      else
         return(sinh(rho*sqrt(1.0-x*x))/sinh(rho));
   }
}


int main(void)
{
   int i,n;
   float rhof,*s;
   double z,eps,a,b,isq3;
   double s0,s1,s2,s3,d,dmax;
   double kp,km,pp,pm;
   double *f;

   printf("\n");
   printf("Check of the distribution produced by random_su2\n");
   printf("------------------------------------------------\n");

   for (;;)
   {
      printf("\n");      
      printf("Specify number of trials (0 exits): ");
      scanf("%d",&n);
      printf("\n");

      if (n<=0)
         exit(0);
            
      printf("Specify rho: ");
      scanf("%lf",&rho);
      printf("\n");      

      f=amalloc(n*sizeof(double),3);
      s=amalloc(4*n*sizeof(float),3);
      rhof=(float)(rho);
   
      for (i=0;i<n;i++)
         random_su2(rhof,&s[4*i]);

      if (rho<=1.0)
         z=1.57;
      else
         z=1.25/(sqrt(rho)*rho);
      eps=1.0e-6*z;

      if (rho<=1.0)
         a=-1.0;
      else
      {
         a=log(rho*eps)/rho;
         if ((a<=1.0)&&(a>=-1.0))
            a=1.0+a;
         else
            a=-1.0;
      }
      b=1.0;
      
      z=cheby_int(a,b,prob1,512,eps);

      dmax=0.0;

      for (i=0;i<n;i++)
      {
         s0=(double)(s[4*i]);
         s1=(double)(s[4*i+1]);
         s2=(double)(s[4*i+2]);
         s3=(double)(s[4*i+3]);         
            
         d=fabs(s0*s0+s1*s1+s2*s2+s3*s3-1.0);
         if (d>dmax)
            dmax=d;

         if (s0<=a)
            f[i]=0.0;
         else if (s0>=b)
            f[i]=1.0;
         else
            f[i]=cheby_int(a,s0,prob1,512,eps)/z;
      }

      ks_test(n,f,&kp,&km);
      ks_prob(n,kp,km,&pp,&pm);

      printf("Maximal deviation |1-s^2|: %6.2e\n\n",dmax);

      printf("Distribution of s[0] \n");
      printf("Kolmogorov-Smirnov test: K+ = %4.2f, K- = %4.2f\n",kp,km);
      printf("This corresponds to Prob(K+) = %4.2f, Prob(K-) = %4.2f\n",pp,pm);
      printf("\n");

      if (rho<=2.5)
         z=1.57;
      else
         z=1.77/sqrt(rho);
      eps=1.0e-6*z;

      if (rho<=2.5)
         a=-1.0;
      else
      {
         a=log(eps)/rho;
         if ((a<=0.0)&&(a>=-1.0))
            a=-sqrt((-a)*(2.0+a));
         else
            a=-1.0;
      }
      b=-a;
      
      z=cheby_int(a,b,prob2,512,eps);
      isq3=1.0/sqrt(3.0);

      for (i=0;i<n;i++)
      {
         s1=(double)(s[4*i+1]);
         s2=(double)(s[4*i+2]);
         s3=(double)(s[4*i+3]);

         s0=isq3*(s1+s2+s3);
            
         if (s0<=a)
            f[i]=0.0;
         else if (s0>=b)
            f[i]=1.0;
         else
            f[i]=cheby_int(a,s0,prob2,512,eps)/z;
      }

      ks_test(n,f,&kp,&km);
      ks_prob(n,kp,km,&pp,&pm);

      printf("Distribution of s[1]+s[2]+s[3]: \n");
      printf("Kolmogorov-Smirnov test: K+ = %4.2f, K- = %4.2f\n",kp,km);
      printf("This corresponds to Prob(K+) = %4.2f, Prob(K-) = %4.2f\n",pp,pm);
      printf("\n");
      
      afree(f);
      afree(s);
   }
   
   exit(0);
}
