
/*******************************************************************************
*
* File check6.c
*
* Copyright (C) 2005 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Comparison of the output of the ISO C and SSE versions of random_su2
*
*******************************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "random.h"
   
#if (defined SSE)

int main(void)
{
   int i;
   float rho,s[4],r[4];
   double d,dmax;
   FILE *dat=NULL;

   printf("\n");
   printf("Comparison of the ISO C and SSE versions of random_su2\n");
   printf("------------------------------------------------------\n\n");

   rlxs_init(0,123456);
   
   printf("SSE code: reading random SU(2) matrices from file .tmp\n\n");
   fflush(stdout);
      
   dat=fopen(".tmp","rb");

   if (dat==NULL)
   {
      printf("Error: file .tmp does not exist ");
      printf("(first need to run the ISO C program)\n");
      printf("Program aborted\n\n");
      exit(1);
   }

   rho=0.0f;
   dmax=0.0;

   for (i=0;i<100;i++)
   {
      random_su2(rho,s);
      fread(&r[0],sizeof(float),4,dat);
      d=(double)((r[0]-s[0])*(r[0]-s[0])+(r[1]-s[1])*(r[1]-s[1])+
                 (r[2]-s[2])*(r[2]-s[2])+(r[3]-s[3])*(r[3]-s[3]));
      if (d>dmax)
         dmax=d;
   }

   rho=0.1f;

   for (i=0;i<100;i++)
   {
      random_su2(rho,s);
      fread(&r[0],sizeof(float),4,dat);
      d=(double)((r[0]-s[0])*(r[0]-s[0])+(r[1]-s[1])*(r[1]-s[1])+
                 (r[2]-s[2])*(r[2]-s[2])+(r[3]-s[3])*(r[3]-s[3]));      
      if (d>dmax)
         dmax=d;
   }

   rho=1.3f;

   for (i=0;i<100;i++)
   {
      random_su2(rho,s);
      fread(&r[0],sizeof(float),4,dat);
      d=(double)((r[0]-s[0])*(r[0]-s[0])+(r[1]-s[1])*(r[1]-s[1])+
                 (r[2]-s[2])*(r[2]-s[2])+(r[3]-s[3])*(r[3]-s[3]));
      if (d>dmax)
         dmax=d;
   }

   rho=6.0f;

   for (i=0;i<100;i++)
   {
      random_su2(rho,s);
      fread(&r[0],sizeof(float),4,dat);
      d=(double)((r[0]-s[0])*(r[0]-s[0])+(r[1]-s[1])*(r[1]-s[1])+
                 (r[2]-s[2])*(r[2]-s[2])+(r[3]-s[3])*(r[3]-s[3]));
      if (d>dmax)
         dmax=d;
   }   

   dmax=sqrt(dmax);
   printf("||s-r||_max = %.2e\n\n",dmax);  

   remove(".tmp");
   exit(0);
}
   
#else

int main(void)
{
   int i;
   float rho,s[4];
   FILE *dat=NULL;

   printf("\n");
   printf("Comparison of the ISO C and SSE versions of random_su2\n");
   printf("------------------------------------------------------\n\n");

   rlxs_init(0,123456);
   
   printf("ISO C code: writing random SU(2) matrices to file .tmp\n\n");
   fflush(stdout);
      
   dat=fopen(".tmp","wb");

   rho=0.0f;

   for (i=0;i<100;i++)
   {
      random_su2(rho,s);
      fwrite(&s[0],sizeof(float),4,dat);
   }

   rho=0.1f;

   for (i=0;i<100;i++)
   {
      random_su2(rho,s);
      fwrite(&s[0],sizeof(float),4,dat);
   }

   rho=1.3f;

   for (i=0;i<100;i++)
   {
      random_su2(rho,s);
      fwrite(&s[0],sizeof(float),4,dat);
   }

   rho=6.0f;

   for (i=0;i<100;i++)
   {
      random_su2(rho,s);
      fwrite(&s[0],sizeof(float),4,dat);
   }   
   
   exit(0);
}
 
#endif
   

