
/*******************************************************************************
*
* File check5.c
*
* Copyright (C) 2005 Luigi Del Debbio, Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Check of the programs blk_sw_term(), update_swdbgr2swd() and
* chmass_swdbgr2swd()
*
*******************************************************************************/

#define MAIN_PROGRAM

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "mpi.h"
#include "su3.h"
#include "flags.h"
#include "random.h"
#include "start.h"
#include "sw_term.h"
#include "block.h"
#include "global.h"


static int is_on_bnd(block_t *b,int ix)
{
   int vol,*up,*dn;

   vol=(*b).vol;

   if (ix>=vol)
      return 1;

   up=(*b).iup[ix];

   if ((up[0]>=vol)||(up[1]>=vol)||(up[2]>=vol)||(up[3]>=vol))
      return 1;

   dn=(*b).idn[ix];

   if ((dn[0]>=vol)||(dn[1]>=vol)||(dn[2]>=vol)||(dn[3]>=vol))
      return 1;

   return 0;
}


static void random_active(void)
{
   int ix,iy,ibx,iby,mu,n,nb,isw;
   su3_dble *ub;
   block_t *b,*bm;

   b=blk_list(HMC_BLOCKS,&nb,&isw);
   bm=b+nb;
   n=0;

   for (;b<bm;b++)
   {
      ub=(*b).ud[0];
      
      for (ix=0;ix<(*b).vol;ix++)
      {
         ibx=is_on_bnd(b,ix);

         for (mu=0;mu<4;mu++)
         {
            iy=(*b).iup[ix][mu];
            iby=is_on_bnd(b,iy);
         
            if ((ibx==0)||(iby==0))
            {
               if (ix<(*b).volh)
                  random_su3_dble(ub+8*(iy-(*b).volh)+2*mu+1);
               else
                  random_su3_dble(ub+8*(ix-(*b).volh)+2*mu);
            }
         }
      }

      set_blk_flags(HMC_BLOCKS,NEW_ACTIVE_UD,n);
      n+=1;
   }
}


static void grid_sw_term(void)
{
   int n,nb,isw;

   blk_list(HMC_BLOCKS,&nb,&isw);

   for (n=0;n<nb;n++)
      blk_sw_term(n);
}


static double cmp_swd(void)
{
   int ix,iy,nb,isw;
   double *u,*um,*bu,d,dmax;
   block_t *b,*bm;

   dmax=0.0;   
   b=blk_list(HMC_BLOCKS,&nb,&isw);
   bm=b+nb;

   for (;b<bm;b++)
   {
      for (ix=0;ix<(*b).vol;ix++)
      {
         iy=(*b).imb[ix];
      
         bu=(*b).swd[2*ix].u;
         u=swd[2*iy].u;
         um=u+72;

         for(;u<um;u++)
         {
            d=fabs((*bu)-(*u));
            if (d>dmax)
               dmax=d;

            bu+=1;
         }
      }
   }

   return dmax;
}


int main(int argc,char *argv[])
{
   int my_rank,bs[4],nb,isw;
   double d,dmax,k1,k2;
   lat_parms_t lat;
   FILE *flog=NULL,*fin=NULL;

   MPI_Init(&argc,&argv);
   MPI_Comm_rank(MPI_COMM_WORLD,&my_rank);

   if (my_rank==0)
   {
      flog=freopen("check5.log","w",stdout);
      fin=freopen("check5.in","r",stdin);
      
      printf("\n");
      printf("Check of the programs blk_sw_term() and update_swdbgr2swd()\n");
      printf("-----------------------------------------------------------\n\n");

      printf("%dx%dx%dx%d lattice, ",NPROC0*L0,NPROC1*L1,NPROC2*L2,NPROC3*L3);
      printf("%dx%dx%dx%d process grid, ",NPROC0,NPROC1,NPROC2,NPROC3);
      printf("%dx%dx%dx%d local lattice\n\n",L0,L1,L2,L3);

      scanf("bs %d %d %d %d",&bs[0],&bs[1],&bs[2],&bs[3]);      
      fclose(fin);

      printf("bs = %d %d %d %d\n",bs[0],bs[1],bs[2],bs[3]);      
      printf("Random gauge field\n\n");
      fflush(flog);
   }

   MPI_Bcast(bs,4,MPI_INT,0,MPI_COMM_WORLD);
   
   start_ranlux(0,1234);
   geometry();
   alloc_ud();
   alloc_bgr(HMC_BLOCKS,bs);
   blk_list(HMC_BLOCKS,&nb,&isw);

   set_lat_parms(0.0,0.1,1.234);
   random_ud();
   assign_ud2udbgr(HMC_BLOCKS);

   swb_term();
   d=cmp_swd();
   MPI_Reduce(&d,&dmax,1,MPI_DOUBLE,MPI_MAX,0,MPI_COMM_WORLD);

   if (my_rank==0)
   {
      printf("After calling swb_term():\n");
      printf("Max|swd-b.swd| = %.2e\n\n",dmax);
   }   
   
   random_active();
   grid_sw_term();
   swap_active(HMC_BLOCKS);
   sw_term();
   d=cmp_swd();
   MPI_Reduce(&d,&dmax,1,MPI_DOUBLE,MPI_MAX,0,MPI_COMM_WORLD);
   
   if (my_rank==0)
   {
      printf("Comparison of blk_sw_term() with sw_term()\n");
      printf("Max|swd-b.swd|= %.2e\n\n",dmax);
   }
   
   random_active();
   update_swdbgr2swd();
   d=cmp_swd();
   MPI_Reduce(&d,&dmax,1,MPI_DOUBLE,MPI_MAX,0,MPI_COMM_WORLD);

   if (my_rank==0)
   {
      printf("Check of update_swdbgr2swd():\n");      
      printf("Max|swd-b.swd|= %.2e\n",dmax);
   }

   swap_active(HMC_BLOCKS);
   sw_term();
   d=cmp_swd();
   MPI_Reduce(&d,&dmax,1,MPI_DOUBLE,MPI_MAX,0,MPI_COMM_WORLD);

   if (my_rank==0)
      printf("Max|swd-b.swd|= %.2e (comparison with sw_term())\n\n",dmax);

   random_active();
   lat=lat_parms();
   k1=lat.kappa;

   chmass_swdbgr2swd(0.2345);
   lat=lat_parms();
   k2=lat.kappa;
   d=cmp_swd();
   MPI_Reduce(&d,&dmax,1,MPI_DOUBLE,MPI_MAX,0,MPI_COMM_WORLD);

   if (my_rank==0)
   {
      printf("Check of chmass_swdbgr2swd():\n");
      printf("Old kappa = %.1f, new kappa = %.4f (should be 0.2345)\n",k1,k2);
      printf("Max|swd-b.swd|= %.2e\n",dmax);
   }

   swap_active(HMC_BLOCKS);
   sw_term();
   d=cmp_swd();
   MPI_Reduce(&d,&dmax,1,MPI_DOUBLE,MPI_MAX,0,MPI_COMM_WORLD);

   if (my_rank==0)
      printf("Max|swd-b.swd|= %.2e (comparison with sw_term())\n",dmax);

   swap_active(HMC_BLOCKS);
   grid_sw_term();   
   d=cmp_swd();
   MPI_Reduce(&d,&dmax,1,MPI_DOUBLE,MPI_MAX,0,MPI_COMM_WORLD);
   free_blk_swbufs();
   error_chk();

   if (my_rank==0)
   {
      printf("Max|swd-b.swd|= %.2e (check of swb)\n\n",dmax);
      fclose(flog);
   }
   
   MPI_Finalize();
   exit(0);
}
