
/*******************************************************************************
*
* File check5.c
*
* Copyright (C) 2005 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Comparison of the output of the ISO C and SSE versions of the cabmar
* program
*
*******************************************************************************/

#define MAIN_PROGRAM

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "mpi.h"
#include "su3.h"
#include "random.h"
#include "start.h"
#include "update.h"
#include "global.h"

#if ((defined SSE)||(defined SSE2))
static su3 su[80],ru[80];
#else
static su3 su[80];
#endif


static void random_staples(su3 *v)
{
   int i;
   float *r;

   r=(float*)(v);
   gauss(r,18);

   for (i=0;i<18;i++)
      r[i]*=3.0f;
}

   
#if ((defined SSE)||(defined SSE2))

int main(int argc,char *argv[])
{
   int my_rank,k,i,j;
   int tag0,nbf0;   
   float beta[4],r1,r2;
   double d,dmax,dmax_all;
   static su3 u,v,w __attribute__ ((aligned (8)));   
   FILE *flog=NULL,*fdat=NULL;   
   MPI_Status stat;
   
   MPI_Init(&argc,&argv);
   MPI_Comm_rank(MPI_COMM_WORLD,&my_rank);
   
   if (my_rank==0)
   {
      flog=freopen("check5.log2","w",stdout);

      printf("\n");
      printf("Comparison of the ISO C and SSE versions of `cabmar'\n");
      printf("----------------------------------------------------\n\n");

      printf("%dx%dx%dx%d lattice, ",NPROC0*L0,NPROC1*L1,NPROC2*L2,NPROC3*L3);
      printf("%dx%dx%dx%d process grid, ",NPROC0,NPROC1,NPROC2,NPROC3);
      printf("%dx%dx%dx%d local lattice\n\n",L0,L1,L2,L3);
      fflush(flog);
   }    
   
   beta[0]=0.0f;
   beta[1]=0.5f;
   beta[2]=3.0f;
   beta[3]=10.0f;   
   
   start_ranlux(0,123456);
   random_staples(&v);
   
   if (my_rank==0)
   {
      printf("SSE code: reading random SU(3) matrices from file .tmp\n\n");
      fdat=fopen(".tmp","rb");

      error_root(fdat==NULL,1,"main [check5]",
                 "File .tmp not found, first need to run ISO C program");
      fread(su,sizeof(su3),80,fdat);      
   }

   for (k=1;k<NPROC;k++)
   {
      if (my_rank==0)
         fread(ru,sizeof(su3),80,fdat);      

      tag0=mpi_tag();
      nbf0=80*18;
      
      if (my_rank==0)
         MPI_Send(ru,nbf0,MPI_FLOAT,k,tag0,MPI_COMM_WORLD);

      if (my_rank==k)
         MPI_Recv(su,nbf0,MPI_FLOAT,0,tag0,MPI_COMM_WORLD,&stat);
   }
   
   dmax=0.0;

   for (k=0;k<4;k++)
   {
      for (i=0;i<10;i++)
      {
         random_su3(&u);      
         cabmar(beta[k],&u,&v,0);
         w=su[20*k+2*i];

         for (j=0;j<18;j++)
         {
            r1=*((float*)(&u)+j);
            r2=*((float*)(&w)+j);
            d=fabs((double)(r1-r2));
            if (d>dmax)
               dmax=d;
         }

         random_su3(&u);            
         cabmar(beta[k],&u,&v,1);
         w=su[20*k+2*i+1];

         for (j=0;j<18;j++)
         {
            r1=*((float*)(&u)+j);
            r2=*((float*)(&w)+j);
            d=fabs((double)(r1-r2));
            if (d>dmax)
               dmax=d;
         }      
      }
   }

   MPI_Reduce(&dmax,&dmax_all,1,MPI_DOUBLE,MPI_MAX,0,MPI_COMM_WORLD);
 
   if (my_rank==0)
   {    
      printf("max{||u-u'||_oo} = %.2e\n\n",dmax_all);
      remove(".tmp");
      fclose(flog);
   }   

   MPI_Finalize();
   exit(0);
}
   
#else

int main(int argc,char *argv[])
{
   int my_rank,k,i;
   int tag0,nbf0;
   float beta[4];
   su3 u,v;
   FILE *flog=NULL, *fdat=NULL;   
   MPI_Status stat;
   
   MPI_Init(&argc,&argv);
   MPI_Comm_rank(MPI_COMM_WORLD,&my_rank);

   if (my_rank==0)
   {
      flog=freopen("check5.log1","w",stdout);

      printf("\n");
      printf("Comparison of the ISO C and SSE versions of `cabmar'\n");
      printf("----------------------------------------------------\n\n");

      printf("%dx%dx%dx%d lattice, ",NPROC0*L0,NPROC1*L1,NPROC2*L2,NPROC3*L3);
      printf("%dx%dx%dx%d process grid, ",NPROC0,NPROC1,NPROC2,NPROC3);
      printf("%dx%dx%dx%d local lattice\n\n",L0,L1,L2,L3);
      fflush(flog);
   }
   
   beta[0]=0.0f;
   beta[1]=0.5f;
   beta[2]=3.0f;
   beta[3]=10.0f;
   
   start_ranlux(0,123456);
   random_staples(&v);

   if (my_rank==0)
   {
      printf("ISO C code: writing random SU(3) matrices to file .tmp\n\n");
      fdat=fopen(".tmp","wb");
   }
   
   for (k=0;k<4;k++)
   {
      for (i=0;i<10;i++)
      {
         random_su3(&u);      
         cabmar(beta[k],&u,&v,0);
         su[20*k+2*i]=u;
         random_su3(&u);
         cabmar(beta[k],&u,&v,1);
         su[20*k+2*i+1]=u;         
      }
   }

   if (my_rank==0)
      fwrite(su,sizeof(su3),80,fdat);

   for (k=1;k<NPROC;k++)
   {
      tag0=mpi_tag();
      nbf0=80*18;
      
      if (my_rank==k)
         MPI_Send(su,nbf0,MPI_FLOAT,0,tag0,MPI_COMM_WORLD);

      if (my_rank==0)
      {
         MPI_Recv(su,nbf0,MPI_FLOAT,k,tag0,MPI_COMM_WORLD,&stat);
         fwrite(su,sizeof(su3),80,fdat);
      }
   }

   if (my_rank==0)
   {
      printf("Data from all processes written to .tmp\n\n");
      fclose(fdat);
      fclose(flog);
   }
      
   MPI_Finalize();   
   exit(0);
}
 
#endif
