
/*******************************************************************************
*
* File dfl.h
*
* Copyright (C) 2007 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
*******************************************************************************/

#ifndef DFL_H
#define DFL_H

#ifndef SU3_H
#include "su3.h"
#endif

typedef struct
{
   int nb,nbb,Ns;
   complex *zee,*zoo,*zoe,*zeo;
} zop_t;

typedef struct
{
   int nb,nbb,Ns;
   complex_dble *zee,*zoo,*zoe,*zeo;
} zop_dble_t;

#ifndef DFL_C
extern void dfl_Pproj_dble(int k,int l);
extern void dfl_Qproj_dble(int k,int l);
extern double dfl_solve_little(int nmx,double res,complex_dble *v,int *status);
extern void dfl_Lpcnd_dble(int k,int l,int *status);
extern void dfl_RLpcnd_dble(int k,int l,int *status);
extern void dfl_RLpcnd(int k,int l,int *status);
#endif

#ifndef DFL_MODES_C
extern void dfl_modes(int ipr,int *status,int *ifail);
extern void dfl_update(int nmx,int ipr,int *status,int *ifail);
#endif

#ifndef DFL_SUBSPACE_C
extern void dfl_sd2vd(int k,complex_dble *v);
extern void dfl_vd2sd(complex_dble *v,int k);
extern void dfl_sub_vd2sd(complex_dble *v,int k);
extern void dfl_s2v(int k,complex *v);
extern void dfl_v2s(complex *v,int k); 
extern void dfl_sub_v2s(complex *v,int k);
extern void dfl_subspace(spinor **mds);
extern void dfl_global_modes(complex **v);
#endif


#ifndef DFL_SAP_GCR_C
extern double dfl_sap_gcr(int nmx,double res,int k,int l,int *status,
                          int *ifail);
#endif

#ifndef VCOM_C
extern void cpv_int_bnd(complex *v);
extern void cpv_ext_bnd(complex *v);
extern void free_vbufs(void);
#endif

#ifndef VDCOM_C
extern void cpvd_int_bnd(complex_dble *v);
extern void cpvd_ext_bnd(complex_dble *v);
extern void free_vdbufs(void);
#endif

#ifndef VFLDS_C
extern void free_vflds(void);
extern void free_vdflds(void);
extern void alloc_vflds(int n);
extern void alloc_vdflds(int n);
extern complex *vflds_base(int *n,int *nv,int *nvec);
extern complex_dble *vdflds_base(int *n,int *nv,int *nvec);
#endif

#ifndef VGRID_C
extern void vgrid(void);
extern int (*vinn(int *nb))[8];
extern int *vipp(int *nbbe,int *nbbo,int *ofsbbe,int *ofsbbo);
#endif

#ifndef VINIT_C
extern void set_v2zero(int n,complex *v);
extern void set_vd2zero(int n,complex_dble *v);
extern void random_v(int n,complex *v,float sigma);
extern void random_vd(int n,complex_dble *v,double sigma);
extern void assign_v2v(int n,complex *v,complex *w);
extern void assign_v2vd(int n,complex *v,complex_dble *w);
extern void assign_vd2v(int n,complex_dble *v,complex *w);
extern void assign_vd2vd(int n,complex_dble *v,complex_dble *w);
#endif

#ifndef ZGEN_C
extern void gather_sd(int vol,int *imb,spinor_dble *pk,spinor_dble *pl);
extern void apply_U2sd(int vol,int *imb,su3_dble *ud,spinor_dble *pk,
                       spinor_dble *pl);
extern void apply_Udag2sd(int vol,int *imb,su3_dble *ud,spinor_dble *pk,
                          spinor_dble *pl);
extern void (*spinor_prod_gamma[])(int vol,spinor_dble *pk,spinor_dble *pl,
                                   complex_dble *sp);
#endif

#ifndef ZHAT_C
extern void Zhat(complex *v,complex *w);
extern void Zeeinv(complex *v,complex *w);
extern void Zooinv(complex *v,complex *w);                   
extern void Zoe(complex *v,complex *w);
extern void Zeo(complex *v,complex *w);
extern void Znohat(complex *v,complex *w);
#endif

#ifndef ZHAT_DBLE_C
extern void Zhat_dble(complex_dble *v,complex_dble *w);
extern void Zeeinv_dble(complex_dble *v,complex_dble *w);
extern void Zooinv_dble(complex_dble *v,complex_dble *w);                   
extern void Zoe_dble(complex_dble *v,complex_dble *w);
extern void Zeo_dble(complex_dble *v,complex_dble *w);
extern void Znohat_dble(complex_dble *v,complex_dble *w);
#endif

#ifndef ZOPS_C
extern zop_dble_t *zopnohat_dble(void);
extern zop_dble_t *zophat_dble(void);
extern zop_t *zopnohat(void);
extern zop_t *zophat(void);
extern void dfl_little_modes(complex **v,complex **zv,complex_dble **vd,
                             complex_dble **zvd,complex_dble *a);
extern void free_zops(void);
#endif

#endif
