
/*******************************************************************************
*
* File run2.c
*
* Copyright (C) 2007 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Computation of quark propagators in two-flavour QCD
*
* Syntax: run2 -i <filename> 
*
* For usage instructions see the file README.run2
*
*******************************************************************************/

#define MAIN_PROGRAM

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <string.h>
#include "mpi.h"
#include "random.h"
#include "misc.h"
#include "start.h"
#include "linalg.h"
#include "sw_term.h"
#include "dirac.h"
#include "dfl.h"
#include "version.h"
#include "global.h"

static int my_rank,endian;
static int id,pid,bc,nc_first,nc_last;
static int bs_sap[4],nkv_sap,nmr_sap,ncy_sap,nmx;
static int bs_dfl[4],Ns,rinv,rnmr,rncy,rvnmx;
static int vnkv,vnmx,vdnmx,nmx;

static double beta,kappa,csw,rkappa;
static double rvres,vres,vdres,res;

static char log_dir[NAME_SIZE],cnfg_dir[NAME_SIZE],sfld_dir[NAME_SIZE];
static char log_file[NAME_SIZE],log_save[NAME_SIZE];
static char sfld_file[NAME_SIZE],cnfg_file[NAME_SIZE],end_file[NAME_SIZE];
static char nbase[NAME_SIZE],ncnfg[NAME_SIZE];
static FILE *fin=NULL,*flog=NULL,*fend=NULL;

static lat_parms_t lat;
static gcr_parms_t gcr;
static dfl_parms_t dfl;


static void read_infile(int argc,char *argv[])
{
   int ifile;
   
   if (my_rank==0)
   {
      flog=freopen("MPI_Errors","w",stdout);

      ifile=find_opt(argc,argv,"-i");
      endian=endianness();
      
      error_root((ifile==0)||(ifile==(argc-1)),1,"read_infile [run2.c]",
                 "Syntax: run2 -i <filename>");

      error_root(endian==UNKNOWN_ENDIAN,1,"read_infile [run2.c]",
                 "Machine has unknown endianness");
      
      fin=freopen(argv[ifile+1],"r",stdin);
      error_root(fin==NULL,1,"read_infile [run2.c]",
                 "Unable to open input file");

      read_line("id","%d\n",&id);
      read_line("pid","%d\n",&pid);
      read_line("log_dir","%s\n",log_dir);
      read_line("cnfg_dir","%s\n",cnfg_dir);
      read_line("sfld_dir","%s\n",sfld_dir);
      read_line("beta","%lf\n",&beta);
      read_line("kappa","%lf\n",&kappa);
      read_line("csw","%lf\n",&csw);
      read_line("bc","%d\n",&bc);
      read_line("nc_first","%d\n",&nc_first);
      read_line("nc_last","%d\n",&nc_last);

      read_line("bs_sap","%d %d %d %d\n",
                &bs_sap[0],&bs_sap[1],&bs_sap[2],&bs_sap[3]);
      read_line("nkv_sap","%d\n",&nkv_sap);
      read_line("nmr_sap","%d\n",&nmr_sap);      
      read_line("ncy_sap","%d\n",&ncy_sap);
      read_line("nmx","%d\n",&nmx);
      read_line("res","%lf\n",&res);

      read_line("bs_dfl","%d %d %d %d\n",
                &bs_dfl[0],&bs_dfl[1],&bs_dfl[2],&bs_dfl[3]);
      read_line("Ns","%d\n",&Ns);
      read_line("rkappa","%lf\n",&rkappa);
      read_line("rinv","%d\n",&rinv);
      read_line("rnmr","%d\n",&rnmr);
      read_line("rncy","%d\n",&rncy);
      read_line("rvnmx","%d\n",&rvnmx);      
      read_line("rvres","%lf\n",&rvres);
      read_line("vnkv","%d\n",&vnkv);
      read_line("vnmx","%d\n",&vnmx);
      read_line("vres","%lf\n",&vres);
      read_line("vdnmx","%d\n",&vdnmx);
      read_line("vdres","%lf",&vdres);
      
      fclose(fin);
   }

   MPI_Bcast(&endian,1,MPI_INT,0,MPI_COMM_WORLD);
   
   MPI_Bcast(&id,1,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&pid,1,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(log_dir,NAME_SIZE,MPI_CHAR,0,MPI_COMM_WORLD);
   MPI_Bcast(cnfg_dir,NAME_SIZE,MPI_CHAR,0,MPI_COMM_WORLD);
   MPI_Bcast(sfld_dir,NAME_SIZE,MPI_CHAR,0,MPI_COMM_WORLD);
   
   MPI_Bcast(&beta,1,MPI_DOUBLE,0,MPI_COMM_WORLD);
   MPI_Bcast(&kappa,1,MPI_DOUBLE,0,MPI_COMM_WORLD);
   MPI_Bcast(&csw,1,MPI_DOUBLE,0,MPI_COMM_WORLD);
   MPI_Bcast(&bc,1,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&nc_first,1,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&nc_last,1,MPI_INT,0,MPI_COMM_WORLD);
      
   MPI_Bcast(bs_sap,4,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&nkv_sap,1,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&nmr_sap,1,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&ncy_sap,1,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&nmx,1,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&res,1,MPI_DOUBLE,0,MPI_COMM_WORLD);

   MPI_Bcast(bs_dfl,4,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&Ns,1,MPI_INT,0,MPI_COMM_WORLD);   
   MPI_Bcast(&rkappa,1,MPI_DOUBLE,0,MPI_COMM_WORLD);
   MPI_Bcast(&rinv,1,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&rnmr,1,MPI_INT,0,MPI_COMM_WORLD);   
   MPI_Bcast(&rncy,1,MPI_INT,0,MPI_COMM_WORLD);   
   MPI_Bcast(&rvnmx,1,MPI_INT,0,MPI_COMM_WORLD);   
   MPI_Bcast(&rvres,1,MPI_DOUBLE,0,MPI_COMM_WORLD);
   MPI_Bcast(&vnkv,1,MPI_INT,0,MPI_COMM_WORLD);   
   MPI_Bcast(&vnmx,1,MPI_INT,0,MPI_COMM_WORLD);   
   MPI_Bcast(&vres,1,MPI_DOUBLE,0,MPI_COMM_WORLD);
   MPI_Bcast(&vdnmx,1,MPI_INT,0,MPI_COMM_WORLD);   
   MPI_Bcast(&vdres,1,MPI_DOUBLE,0,MPI_COMM_WORLD);
}


static void setup_files(void)
{
   check_dir_root(log_dir);
   check_dir_root(cnfg_dir);
   check_dir_root(sfld_dir);
   
   sprintf(nbase,"%dx%dx%dx%db%1.2fk%.5fc%.5fid%d",
           NPROC0*L0,NPROC1*L1,NPROC2*L2,NPROC3*L3,beta,kappa,csw,id);

   sprintf(ncnfg,"/.%d.log~",pid);   
   error_root((strlen(log_dir)+strlen(nbase)+strlen(ncnfg))>=NAME_SIZE,1,
              "setup_files [run2.c]","log_dir name is too long");
      
   sprintf(log_file,"%s/%s.%d.log",log_dir,nbase,pid);
   sprintf(end_file,"%s/%s.%d.end",log_dir,nbase,pid);
   sprintf(log_save,"%s~",log_file);

   sprintf(ncnfg,"/n%d",nc_last);
   error_root((strlen(cnfg_dir)+strlen(ncnfg))>=NAME_SIZE,1,
              "setup_files [run2.c]","cnfg_dir name is too long");   

   sprintf(ncnfg,"/.%dn%dc43",pid,nc_last);
   error_root((strlen(sfld_dir)+strlen(ncnfg))>=NAME_SIZE,1,
              "setup_files [run2.c]","sfld_dir name is too long");   

   fend=fopen(log_file,"r");
   error_root(fend!=NULL,1,"check_files [run2.c]",
              "Attempt to overwrite old *.log");
}


static void print_info(void)
{
   int ip;
   
   if (my_rank==0)
   {
      ip=ftell(flog);
      fclose(flog);

      if (ip==0)
         remove("MPI_Errors");
      
      flog=freopen(log_file,"w",stdout);
      error_root(flog==NULL,1,"print_info [run2.c]","Unable to open log file");
      
      printf("\n");
      printf("Computation of quark propagators\n");
      printf("--------------------------------\n\n");

      printf("Program version %s\n",DD_HMC_RELEASE);

      if (endian==LITTLE_ENDIAN)
         printf("The machine is little endian\n");
      else
         printf("The machine is big endian\n");
         
      printf("\n");      
      printf("%dx%dx%dx%d lattice, ",NPROC0*L0,NPROC1*L1,NPROC2*L2,NPROC3*L3);
      printf("%dx%dx%dx%d process grid, ",NPROC0,NPROC1,NPROC2,NPROC3);
      printf("%dx%dx%dx%d local lattice\n\n",L0,L1,L2,L3);

      printf("O(a) improved theory\n");
      printf("beta = %.6f\n",lat.beta);
      printf("kappa = %.6f\n",lat.kappa);
      printf("csw = %.6f\n",lat.csw);
      printf("bc = %+d\n\n",bc);
      
      printf("bs_sap = %d %d %d %d\n",
             bs_sap[0],bs_sap[1],bs_sap[2],bs_sap[3]);
      printf("nkv_sap,nmr_sap,ncy_sap = %d,%d,%d\n",
             gcr.nkv,gcr.nmr,gcr.ncy);
      printf("nmx = %d\n",nmx);
      printf("res = %.1e\n\n",res);   

      printf("bs_dfl = %d %d %d %d\n",
             bs_dfl[0],bs_dfl[1],bs_dfl[2],bs_dfl[3]);
      printf("Ns = %d\n",dfl.Ns);
      printf("rkappa = %.6f\n",dfl.rkappa);
      printf("rinv = %d\n",dfl.rinv);
      printf("rnmr = %d\n",dfl.rnmr);
      printf("rncy = %d\n",dfl.rncy);         
      printf("rvnmx = %d\n",dfl.rvnmx);
      printf("rvres = %.1e\n",dfl.rvres);
      printf("vnkv = %d\n",dfl.vnkv);         
      printf("vnmx = %d\n",dfl.vnmx);
      printf("vres = %.1e\n",dfl.vres);
      printf("vdnmx = %d\n",dfl.vdnmx);
      printf("vdres = %.1e\n\n",dfl.vdres);

      printf("Configurations no %d -> %d\n",nc_first,nc_last);
      fflush(flog);
   }
}


static void srcfld(int idirac,int icolour,spinor_dble *sd)
{
   int x[4],ip,ix;
   double *s;
   
   set_sd2zero(VOLUME,sd);

   x[0]=0;
   x[1]=0;
   x[2]=0;
   x[3]=0;
   
   ipt_global(x,&ip,&ix);

   if (my_rank==ip)
   {
      s=(double*)(sd+ix);
      s[6*(idirac-1)+2*(icolour-1)]=1.0;
   }
}


static double residue(spinor_dble *s,spinor_dble *r)
{
   int ir,iw;
   double d;
   complex_dble z;
   spinor_dble **wsd;
   
   wsd=reserve_wsd(1);

   ir=(r-psd[0][0])/NSPIN;   
   iw=(wsd[0]-psd[0][0])/NSPIN;

   Qnohat_dble(ir,iw);
   mulg5_dble(VOLUME,wsd[0]);
   z.re=-1.0;
   z.im=0.0;
   mulc_spinor_add_dble(VOLUME,wsd[0],s,z);
   d=norm_square_dble(VOLUME,1,wsd[0]);

   release_wsd();

   return sqrt(d);
}


static void check_endflag(int *iend)
{
   if (my_rank==0)
   {
      fend=fopen(end_file,"r");

      if (fend!=NULL)
      {
         fclose(fend);
         remove(end_file);
         (*iend)=1;
         printf("\nEnd flag set, run stopped\n");
      }
   }

   MPI_Bcast(iend,1,MPI_INT,0,MPI_COMM_WORLD);
}


int main(int argc,char *argv[])
{
   int icnfg,idirac,icolour,status[3],ifail;
   int iend,iw0,iw1;
   double cres,wt1,wt2,wtavg;
   spinor_dble **wsd;

   MPI_Init(&argc,&argv);
   MPI_Comm_rank(MPI_COMM_WORLD,&my_rank);

   read_infile(argc,argv);
   setup_files();

   geometry();
   alloc_ud();
   alloc_swd();
   alloc_u();
   alloc_sw();
   if (Ns<(2*nkv_sap))
      alloc_s(2*nkv_sap+1);
   else
      alloc_s(Ns+2);
   alloc_sd(5);
   
   lat=set_lat_parms(beta,kappa,csw);
   gcr=set_gcr_parms(bs_sap,nmr_sap,ncy_sap,nkv_sap);
   dfl=set_dfl_parms(bs_dfl,Ns,rinv,rnmr,rncy,rvnmx,vnkv,vnmx,vdnmx,
                     rkappa,rvres,vres,vdres,0,0);

   print_info();

   wsd=reserve_wsd(2);
   iw0=(wsd[0]-psd[0][0])/NSPIN;
   iw1=(wsd[1]-psd[0][0])/NSPIN;
   iend=0;
   wtavg=0.0;

   for (icnfg=nc_first;(iend==0)&&(icnfg<=nc_last);icnfg++)
   {
      MPI_Barrier(MPI_COMM_WORLD);
      wt1=MPI_Wtime();      

      message("\nConfiguration no %d:\n",icnfg);
      sprintf(cnfg_file,"%s/%sn%d",cnfg_dir,nbase,icnfg);
      import_cnfg(cnfg_file);
      if (bc==-1)
         flipbcd();

      sw_term();
      assign_ud2u();
      assign_swd2sw();
      assign_ud2ubgr(GCR_BLOCKS);

      dfl_modes(0,status,&ifail);
      message("Deflation subspace generation: status = %d, ifail = %d\n",
              status[0],ifail);

      error_root(ifail!=0,1,"main [run2.c]","Subspace generation failed");
      
      for (idirac=1;idirac<=4;idirac++)
      {
         for (icolour=1;icolour<=3;icolour++)
         {
            srcfld(idirac,icolour,wsd[0]);         
            dfl_sap_gcr(nmx,res,iw0,iw1,status,&ifail);
            cres=residue(wsd[0],wsd[1]);

            message("(%d%d)-component: status = %3d (little: %3d,%3d), "
                    "ifail = %d, check = %.1e\n",idirac,icolour,status[0],
                    status[1],status[2],ifail,cres);

            error_root((status[0]<0)||(ifail!=0),1,"main [run2.c]",
                       "Propagator computation failed");
            
            sprintf(sfld_file,"%s/%s.%dn%dc%d%d",
                    sfld_dir,nbase,pid,icnfg,idirac,icolour);
            export_sfld(sfld_file,iw1);
         }
      }
      
      MPI_Barrier(MPI_COMM_WORLD);
      wt2=MPI_Wtime();

      message("Configuration fully processed in %.2e sec\n",wt2-wt1);

      wtavg+=(wt2-wt1);
      check_endflag(&iend);
      error_chk();

      if (my_rank==0)
      {
         if ((iend==1)||(icnfg==nc_last))
         {
            printf("\n");
            printf("Processed configurations no %d -> %d\n",nc_first,icnfg);
            printf("Average time per configuration = %.2e sec\n\n",
                   wtavg/(double)(icnfg-nc_first+1));
         }
         
         fflush(flog);
         copy_file(log_file,log_save);
      }
   }

   if (my_rank==0)
      fclose(flog);

   MPI_Finalize();
   exit(0);
}
