
/*******************************************************************************
*
* File mvfld.c
*
* Copyright (C) 2006 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Random translation of the double-precision gauge field
*
* The externally accessible function is
*
*   int mvfld(void)
*     Translates the global double-precision gauge field by a random
*     vector. On exit the program returns the number of elementary
*     shifts along the lattice axes which were performed
*
* Notes:
*
* The program obtains the random vectors from the block shift generator
* [see hmc/bshift.c]. The latter is (re)started when mvfld() is first
* called, using the block size of the current HMC_BLOCKS grid. An error
* occurs if the grid is not allocated at this point.
*
* If the block size of HMC_BLOCKS grid is changed at a later stage, the
* generator is restarted.
*
* This programs performs global operations and should therefore be called
* simultaneously on all processes. The translations are carried out by the
* program shift_ud() [update/shift.c].
*
*******************************************************************************/

#define MVFLD_C

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "mpi.h"
#include "start.h"
#include "update.h"
#include "hmc.h"

static int bs[4]={0,0,0,0};


int mvfld(void)
{
   int nb,isw,*bs0;
   int mu,ns,svec[4];
   block_t *b0;
   
   b0=blk_list(HMC_BLOCKS,&nb,&isw);
   error(b0==NULL,1,"shift_field [mvfld.c]",
         "The HMC_BLOCKS grid is not allocated");
   bs0=(*b0).bs;
   
   if ((bs[0]!=bs0[0])||(bs[1]!=bs0[1])||(bs[2]!=bs0[2])||(bs[3]!=bs0[3]))
   {
      bs[0]=bs0[0];
      bs[1]=bs0[1];
      bs[2]=bs0[2];
      bs[3]=bs0[3];

      start_bshift(bs);      
   }
   
   next_bshift(svec);

   for (mu=0;mu<4;mu++)
   {
      if (svec[mu]>(bs[mu]/2))
         svec[mu]-=bs[mu];
   }

   ns=shift_ud(svec);

   return ns;
}
