
/*******************************************************************************
*
* File mutils.c
*
* Copyright (C) 2004 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)

* Utility functions used in main programs
*
* The externally accessible functions are
*
*   int find_opt(int argc,char *argv[],char *opt)
*     Returns the position of the option opt in the array argv[]. Only
*     the elements argv[1],..,argv[argc-1] are scanned and 0 is returned
*     if opt is not found
*
*   int digits(double x,double dx,char *fmt)
*     Assuming x is a value with error dx, this program returns the number n
*     of fractional digits to print so that all significant digits plus two
*     more are shown. The print format fmt has to be "e" or "f" depending on
*     whether the number is to be printed using the "%.ne" or "%.nf" format
*     string. In the second case dx has to be in the range 0<dx<1, and
*     (int)(10^n*dx) is then a two-digit integer that represents the error
*     in the last two digits of the printed value
*
*******************************************************************************/

#define MUTILS_C

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include <float.h>
#include "start.h"


int find_opt(int argc,char *argv[],char *opt)
{
   int k;

   for (k=1;k<argc;k++)
      if (strcmp(argv[k],opt)==0)
         return k;

   return 0;
}


int digits(double x,double dx,char *fmt)
{
   error(dx<0.0,1,"digits [misc.c]","Improper input data (negative error)");
   
   if (strcmp(fmt,"e")==0)
   {
      if (dx==0.0)
         return DBL_DIG;
      else if (dx>=fabs(x))
         return 1;
      else
         return (int)(floor(1.0+log10(fabs(x)))-floor(log10(dx)));
   }
   else if (strcmp(fmt,"f")==0)
   {
      error((dx==0.0)||(dx>=1.0),1,"digits [misc.c]",
            "Improper input data (error out of range for fixed format)");

      return (int)(1.0-floor(log10(dx)));
   }
   else
      error(1,1,"digits [misc.c]","Unknown data format");

   return 0;
}

