
/*******************************************************************************
*
* File staples.c
*
* Copyright (C) 2005 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Computation of the "staples"
*
* The externally accessible function is
*
*   void staples(int ix,int mu,su3 *v)
*     Computes the 3x3 matrix v such that the part of the Wilson action
*     that depends on u=U(x,mu) is given by -2/g_0^2*Re[tr{u*v^dagger}]
*
*******************************************************************************/

#define STAPLES_C

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "su3.h"
#include "global.h"

#if (defined SSE)

#include "sse.h"

static su3 *u1up,*u2up,*u3up;
static su3 *u1dn,*u2dn,*u3dn;
static sse_float v1,v2,v3;
static sse_float x1,x2,x3;


static void sse_staple(void)
{
   
/*
 * Computation of the upper staple
 */

   __asm__ __volatile__ ("movlps %0, %%xmm0 \n\t"
                         "movlps %1, %%xmm1 \n\t"
                         "movlps %2, %%xmm2 \n\t"
                         "movhps %3, %%xmm0 \n\t"
                         "movhps %4, %%xmm1 \n\t"
                         "movhps %5, %%xmm2 \n\t"
                         "mulps %6, %%xmm0 \n\t"
                         "mulps %6, %%xmm1 \n\t"
                         "mulps %6, %%xmm2"                         
                         :
                         :
                         "m" ((*u3up).c11),
                         "m" ((*u3up).c12),
                         "m" ((*u3up).c13),
                         "m" ((*u3up).c21),
                         "m" ((*u3up).c22),
                         "m" ((*u3up).c23),
                         "m" (_sse_sgn24)
                         :
                         "xmm0", "xmm1", "xmm2");
   
   _sse_su3_multiply((*u2up));

   __asm__ __volatile__ ("movaps %%xmm3, %%xmm0 \n\t"
                         "movaps %%xmm4, %%xmm1 \n\t"
                         "movaps %%xmm5, %%xmm2"
                         :
                         :
                         :
                         "xmm0", "xmm1", "xmm2");

   _sse_su3_multiply((*u1up));

/*
 * Add to v
 */   

   __asm__ __volatile__ ("movaps %%xmm3, %%xmm0 \n\t"
                         "movaps %%xmm4, %%xmm1 \n\t"
                         "movaps %%xmm5, %%xmm2 \n\t"
                         "shufps $0x1e, %%xmm0, %%xmm0 \n\t"
                         "shufps $0x1e, %%xmm1, %%xmm1 \n\t"
                         "shufps $0x1e, %%xmm2, %%xmm2 \n\t"
                         "movaps %%xmm0, %%xmm6 \n\t"
                         "movaps %%xmm1, %%xmm7 \n\t"
                         "mulps %%xmm4, %%xmm0 \n\t"
                         "mulps %%xmm5, %%xmm6 \n\t"                         
                         "mulps %%xmm3, %%xmm1 \n\t"
                         "mulps %%xmm5, %%xmm7 \n\t"
                         "subps %%xmm0, %%xmm1 \n\t"                         
                         "movaps %%xmm2, %%xmm0 \n\t"
                         "mulps %%xmm4, %%xmm2 \n\t"
                         "mulps %%xmm3, %%xmm0 \n\t"
                         "subps %%xmm7, %%xmm2 \n\t"
                         "subps %%xmm0, %%xmm6"
                         :
                         :
                         :
                         "xmm0", "xmm1", "xmm2", "xmm6", "xmm7");

   __asm__ __volatile__ ("addps %0, %%xmm1 \n\t"
                         "addps %1, %%xmm2 \n\t"
                         "addps %2, %%xmm6 \n\t"
                         "addps %3, %%xmm3 \n\t"
                         "addps %4, %%xmm4 \n\t"
                         "addps %5, %%xmm5"
                         :
                         :
                         "m" (x3),
                         "m" (x1),
                         "m" (x2),
                         "m" (v1),
                         "m" (v2),
                         "m" (v3)
                         :
                         "xmm1", "xmm2", "xmm3", "xmm4", "xmm5", "xmm6");

   __asm__ __volatile__ ("movaps %%xmm1, %0 \n\t"
                         "movaps %%xmm2, %1 \n\t"
                         "movaps %%xmm6, %2 \n\t"
                         "movaps %%xmm3, %3 \n\t"
                         "movaps %%xmm4, %4 \n\t"
                         "movaps %%xmm5, %5"                         
                         :
                         "=m" (x3),
                         "=m" (x1),
                         "=m" (x2),
                         "=m" (v1),
                         "=m" (v2),
                         "=m" (v3));
   
/*
 * Computation of the lower staple
 */ 

   __asm__ __volatile__ ("movups %0, %%xmm0 \n\t"
                         "movups %1, %%xmm1 \n\t"
                         "movups %2, %%xmm2"
                         :
                         :
                         "m" ((*u3dn).c11),
                         "m" ((*u3dn).c21),
                         "m" ((*u3dn).c31)
                         :
                         "xmm0", "xmm1", "xmm2");

   _sse_su3_multiply((*u2dn));

   __asm__ __volatile__ ("movaps %%xmm3, %%xmm0 \n\t"
                         "movaps %%xmm4, %%xmm1 \n\t"
                         "movaps %%xmm5, %%xmm2"
                         :
                         :
                         :
                         "xmm0", "xmm1", "xmm2");

   _sse_su3_inverse_multiply((*u1dn));

/*
 * Add to v
 */

   __asm__ __volatile__ ("movaps %%xmm3, %%xmm0 \n\t"
                         "movaps %%xmm4, %%xmm1 \n\t"
                         "movaps %%xmm5, %%xmm2 \n\t"
                         "shufps $0x1e, %%xmm0, %%xmm0 \n\t"
                         "shufps $0x1e, %%xmm1, %%xmm1 \n\t"
                         "shufps $0x1e, %%xmm2, %%xmm2 \n\t"
                         "movaps %%xmm0, %%xmm6 \n\t"
                         "movaps %%xmm1, %%xmm7 \n\t"
                         "mulps %%xmm4, %%xmm0 \n\t"
                         "mulps %%xmm5, %%xmm6 \n\t"                         
                         "mulps %%xmm3, %%xmm1 \n\t"
                         "mulps %%xmm5, %%xmm7 \n\t"
                         "subps %%xmm0, %%xmm1 \n\t"                         
                         "movaps %%xmm2, %%xmm0 \n\t"
                         "mulps %%xmm4, %%xmm2 \n\t"
                         "mulps %%xmm3, %%xmm0 \n\t"
                         "subps %%xmm7, %%xmm2 \n\t"
                         "subps %%xmm0, %%xmm6 \n\t"
                         :
                         :
                         :
                         "xmm0", "xmm1", "xmm2", "xmm6", "xmm7");
   
   __asm__ __volatile__ ("addps %0, %%xmm1 \n\t"
                         "addps %1, %%xmm2 \n\t"
                         "addps %2, %%xmm6 \n\t"
                         "addps %3, %%xmm3 \n\t"
                         "addps %4, %%xmm4 \n\t"
                         "addps %5, %%xmm5"                         
                         :
                         :
                         "m" (x3),
                         "m" (x1),
                         "m" (x2),
                         "m" (v1),
                         "m" (v2),
                         "m" (v3)
                         :
                         "xmm1", "xmm2", "xmm3", "xmm4", "xmm5", "xmm6");

   __asm__ __volatile__ ("movaps %%xmm1, %0 \n\t"
                         "movaps %%xmm2, %1 \n\t"
                         "movaps %%xmm6, %2 \n\t"
                         "movaps %%xmm3, %3 \n\t"
                         "movaps %%xmm4, %4 \n\t"
                         "movaps %%xmm5, %5"
                         :
                         "=m" (x3),
                         "=m" (x1),
                         "=m" (x2),                       
                         "=m" (v1),
                         "=m" (v2),
                         "=m" (v3));
}


void staples(int ix,int mu,su3 *v)
{
   int i,nu,ixpmu,ixpnu,ixmnu,ixpmumnu;

   __asm__ __volatile__ ("xorps %%xmm3, %%xmm3 \n\t"
                         "xorps %%xmm4, %%xmm4 \n\t"
                         "xorps %%xmm5, %%xmm5 \n\t"
                         "movaps %%xmm3, %0 \n\t"
                         "movaps %%xmm4, %1 \n\t"
                         "movaps %%xmm5, %2 \n\t"
                         "movaps %%xmm3, %3 \n\t"
                         "movaps %%xmm4, %4 \n\t"
                         "movaps %%xmm5, %5"
                         :
                         "=m" (x3),
                         "=m" (x1),
                         "=m" (x2),                       
                         "=m" (v1),
                         "=m" (v2),
                         "=m" (v3)
                         :
                         :
                         "xmm3", "xmm4", "xmm5");                         

   ixpmu=iup[ix][mu];

   for (i=1;i<4;i++)
   {
      nu=(mu+i)&0x3;
      ixpnu=iup[ix][nu];
      ixmnu=idn[ix][nu];
      ixpmumnu=idn[ixpmu][nu];
      
      u1up=pu[ix][nu];
      u2up=pu[ixpnu][mu];
      u3up=pu[ixpmu][nu];   
      
      u1dn=pu[ixmnu][nu];
      u2dn=pu[ixmnu][mu];
      u3dn=pu[ixpmumnu][nu];

      sse_staple();
   }

   __asm__ __volatile__ ("movaps %3, %%xmm0 \n\t"
                         "movaps %4, %%xmm1 \n\t"
                         "movaps %5, %%xmm2 \n\t"
                         "movups %%xmm0, %0 \n\t"
                         "movups %%xmm1, %1 \n\t"
                         "movups %%xmm2, %2"
                         :
                         "=m" ((*v).c11),
                         "=m" ((*v).c21),
                         "=m" ((*v).c31)
                         :
                         "m" (v1),
                         "m" (v2),
                         "m" (v3)
                         :
                         "xmm0", "xmm1", "xmm2");
   
   (*v).c13.re=x1.c1-x1.c2;
   (*v).c13.im=x1.c3+x1.c4;

   (*v).c23.re=x2.c1-x2.c2;
   (*v).c23.im=x2.c3+x2.c4;

   (*v).c33.re=x3.c1-x3.c2;
   (*v).c33.im=x3.c3+x3.c4;
}

#else

static su3_vector psi1,psi2,chi1,chi2;
static su3 *u1up,*u2up,*u3up;
static su3 *u1dn,*u2dn,*u3dn;
static const su3 v0={{0.0f,0.0f},{0.0f,0.0f},{0.0f,0.0f},
                     {0.0f,0.0f},{0.0f,0.0f},{0.0f,0.0f},
                     {0.0f,0.0f},{0.0f,0.0f},{0.0f,0.0f}};


static void add_to_v(su3 *v)
{
   (*v).c11.re+=psi1.c1.re;
   (*v).c11.im+=psi1.c1.im;
   (*v).c21.re+=psi1.c2.re;
   (*v).c21.im+=psi1.c2.im;
   (*v).c31.re+=psi1.c3.re;
   (*v).c31.im+=psi1.c3.im;

   (*v).c12.re+=psi2.c1.re;
   (*v).c12.im+=psi2.c1.im;
   (*v).c22.re+=psi2.c2.re;
   (*v).c22.im+=psi2.c2.im;
   (*v).c32.re+=psi2.c3.re;
   (*v).c32.im+=psi2.c3.im;

   (*v).c13.re+=
      (psi1.c2.re*psi2.c3.re-psi1.c2.im*psi2.c3.im)-
      (psi2.c2.re*psi1.c3.re-psi2.c2.im*psi1.c3.im);
   (*v).c13.im+=
      (psi2.c2.re*psi1.c3.im+psi2.c2.im*psi1.c3.re)-
      (psi1.c2.re*psi2.c3.im+psi1.c2.im*psi2.c3.re);
   (*v).c23.re+=
      (psi1.c3.re*psi2.c1.re-psi1.c3.im*psi2.c1.im)-
      (psi2.c3.re*psi1.c1.re-psi2.c3.im*psi1.c1.im);
   (*v).c23.im+=
      (psi2.c3.re*psi1.c1.im+psi2.c3.im*psi1.c1.re)-
      (psi1.c3.re*psi2.c1.im+psi1.c3.im*psi2.c1.re);
   (*v).c33.re+=
      (psi1.c1.re*psi2.c2.re-psi1.c1.im*psi2.c2.im)-
      (psi2.c1.re*psi1.c2.re-psi2.c1.im*psi1.c2.im);
   (*v).c33.im+=
      (psi2.c1.re*psi1.c2.im+psi2.c1.im*psi1.c2.re)-
      (psi1.c1.re*psi2.c2.im+psi1.c1.im*psi2.c2.re);
}


static void up_staple(void)
{
   chi1.c1.re=
      (*u2up).c11.re*(*u3up).c11.re+(*u2up).c11.im*(*u3up).c11.im+
      (*u2up).c12.re*(*u3up).c12.re+(*u2up).c12.im*(*u3up).c12.im+ 
      (*u2up).c13.re*(*u3up).c13.re+(*u2up).c13.im*(*u3up).c13.im;
   chi1.c1.im=
      (*u2up).c11.im*(*u3up).c11.re-(*u2up).c11.re*(*u3up).c11.im+
      (*u2up).c12.im*(*u3up).c12.re-(*u2up).c12.re*(*u3up).c12.im+ 
      (*u2up).c13.im*(*u3up).c13.re-(*u2up).c13.re*(*u3up).c13.im;
   chi1.c2.re=
      (*u2up).c21.re*(*u3up).c11.re+(*u2up).c21.im*(*u3up).c11.im+ 
      (*u2up).c22.re*(*u3up).c12.re+(*u2up).c22.im*(*u3up).c12.im+
      (*u2up).c23.re*(*u3up).c13.re+(*u2up).c23.im*(*u3up).c13.im;
   chi1.c2.im=
      (*u2up).c21.im*(*u3up).c11.re-(*u2up).c21.re*(*u3up).c11.im+
      (*u2up).c22.im*(*u3up).c12.re-(*u2up).c22.re*(*u3up).c12.im+
      (*u2up).c23.im*(*u3up).c13.re-(*u2up).c23.re*(*u3up).c13.im;
   chi1.c3.re=
      (*u2up).c31.re*(*u3up).c11.re+(*u2up).c31.im*(*u3up).c11.im+
      (*u2up).c32.re*(*u3up).c12.re+(*u2up).c32.im*(*u3up).c12.im+ 
      (*u2up).c33.re*(*u3up).c13.re+(*u2up).c33.im*(*u3up).c13.im;
   chi1.c3.im=
      (*u2up).c31.im*(*u3up).c11.re-(*u2up).c31.re*(*u3up).c11.im+ 
      (*u2up).c32.im*(*u3up).c12.re-(*u2up).c32.re*(*u3up).c12.im+
      (*u2up).c33.im*(*u3up).c13.re-(*u2up).c33.re*(*u3up).c13.im;

   chi2.c1.re=
      (*u2up).c11.re*(*u3up).c21.re+(*u2up).c11.im*(*u3up).c21.im+
      (*u2up).c12.re*(*u3up).c22.re+(*u2up).c12.im*(*u3up).c22.im+ 
      (*u2up).c13.re*(*u3up).c23.re+(*u2up).c13.im*(*u3up).c23.im;
   chi2.c1.im=
      (*u2up).c11.im*(*u3up).c21.re-(*u2up).c11.re*(*u3up).c21.im+
      (*u2up).c12.im*(*u3up).c22.re-(*u2up).c12.re*(*u3up).c22.im+ 
      (*u2up).c13.im*(*u3up).c23.re-(*u2up).c13.re*(*u3up).c23.im;
   chi2.c2.re=
      (*u2up).c21.re*(*u3up).c21.re+(*u2up).c21.im*(*u3up).c21.im+ 
      (*u2up).c22.re*(*u3up).c22.re+(*u2up).c22.im*(*u3up).c22.im+
      (*u2up).c23.re*(*u3up).c23.re+(*u2up).c23.im*(*u3up).c23.im;
   chi2.c2.im=
      (*u2up).c21.im*(*u3up).c21.re-(*u2up).c21.re*(*u3up).c21.im+
      (*u2up).c22.im*(*u3up).c22.re-(*u2up).c22.re*(*u3up).c22.im+
      (*u2up).c23.im*(*u3up).c23.re-(*u2up).c23.re*(*u3up).c23.im;
   chi2.c3.re=
      (*u2up).c31.re*(*u3up).c21.re+(*u2up).c31.im*(*u3up).c21.im+
      (*u2up).c32.re*(*u3up).c22.re+(*u2up).c32.im*(*u3up).c22.im+ 
      (*u2up).c33.re*(*u3up).c23.re+(*u2up).c33.im*(*u3up).c23.im;
   chi2.c3.im=
      (*u2up).c31.im*(*u3up).c21.re-(*u2up).c31.re*(*u3up).c21.im+ 
      (*u2up).c32.im*(*u3up).c22.re-(*u2up).c32.re*(*u3up).c22.im+
      (*u2up).c33.im*(*u3up).c23.re-(*u2up).c33.re*(*u3up).c23.im;

   _su3_multiply(psi1,(*u1up),chi1);
   _su3_multiply(psi2,(*u1up),chi2);   
}


static void dn_staple(void)
{
   chi1.c1.re=
      (*u2dn).c11.re*(*u3dn).c11.re-(*u2dn).c11.im*(*u3dn).c11.im+
      (*u2dn).c12.re*(*u3dn).c21.re-(*u2dn).c12.im*(*u3dn).c21.im+ 
      (*u2dn).c13.re*(*u3dn).c31.re-(*u2dn).c13.im*(*u3dn).c31.im;
   chi1.c1.im=
      (*u2dn).c11.im*(*u3dn).c11.re+(*u2dn).c11.re*(*u3dn).c11.im+
      (*u2dn).c12.im*(*u3dn).c21.re+(*u2dn).c12.re*(*u3dn).c21.im+ 
      (*u2dn).c13.im*(*u3dn).c31.re+(*u2dn).c13.re*(*u3dn).c31.im;
   chi1.c2.re=
      (*u2dn).c21.re*(*u3dn).c11.re-(*u2dn).c21.im*(*u3dn).c11.im+ 
      (*u2dn).c22.re*(*u3dn).c21.re-(*u2dn).c22.im*(*u3dn).c21.im+
      (*u2dn).c23.re*(*u3dn).c31.re-(*u2dn).c23.im*(*u3dn).c31.im;
   chi1.c2.im=
      (*u2dn).c21.im*(*u3dn).c11.re+(*u2dn).c21.re*(*u3dn).c11.im+
      (*u2dn).c22.im*(*u3dn).c21.re+(*u2dn).c22.re*(*u3dn).c21.im+
      (*u2dn).c23.im*(*u3dn).c31.re+(*u2dn).c23.re*(*u3dn).c31.im;
   chi1.c3.re=
      (*u2dn).c31.re*(*u3dn).c11.re-(*u2dn).c31.im*(*u3dn).c11.im+
      (*u2dn).c32.re*(*u3dn).c21.re-(*u2dn).c32.im*(*u3dn).c21.im+ 
      (*u2dn).c33.re*(*u3dn).c31.re-(*u2dn).c33.im*(*u3dn).c31.im;
   chi1.c3.im=
      (*u2dn).c31.im*(*u3dn).c11.re+(*u2dn).c31.re*(*u3dn).c11.im+ 
      (*u2dn).c32.im*(*u3dn).c21.re+(*u2dn).c32.re*(*u3dn).c21.im+
      (*u2dn).c33.im*(*u3dn).c31.re+(*u2dn).c33.re*(*u3dn).c31.im;

   chi2.c1.re=
      (*u2dn).c11.re*(*u3dn).c12.re-(*u2dn).c11.im*(*u3dn).c12.im+
      (*u2dn).c12.re*(*u3dn).c22.re-(*u2dn).c12.im*(*u3dn).c22.im+ 
      (*u2dn).c13.re*(*u3dn).c32.re-(*u2dn).c13.im*(*u3dn).c32.im;
   chi2.c1.im=
      (*u2dn).c11.im*(*u3dn).c12.re+(*u2dn).c11.re*(*u3dn).c12.im+
      (*u2dn).c12.im*(*u3dn).c22.re+(*u2dn).c12.re*(*u3dn).c22.im+ 
      (*u2dn).c13.im*(*u3dn).c32.re+(*u2dn).c13.re*(*u3dn).c32.im;
   chi2.c2.re=
      (*u2dn).c21.re*(*u3dn).c12.re-(*u2dn).c21.im*(*u3dn).c12.im+ 
      (*u2dn).c22.re*(*u3dn).c22.re-(*u2dn).c22.im*(*u3dn).c22.im+
      (*u2dn).c23.re*(*u3dn).c32.re-(*u2dn).c23.im*(*u3dn).c32.im;
   chi2.c2.im=
      (*u2dn).c21.im*(*u3dn).c12.re+(*u2dn).c21.re*(*u3dn).c12.im+
      (*u2dn).c22.im*(*u3dn).c22.re+(*u2dn).c22.re*(*u3dn).c22.im+
      (*u2dn).c23.im*(*u3dn).c32.re+(*u2dn).c23.re*(*u3dn).c32.im;
   chi2.c3.re=
      (*u2dn).c31.re*(*u3dn).c12.re-(*u2dn).c31.im*(*u3dn).c12.im+
      (*u2dn).c32.re*(*u3dn).c22.re-(*u2dn).c32.im*(*u3dn).c22.im+ 
      (*u2dn).c33.re*(*u3dn).c32.re-(*u2dn).c33.im*(*u3dn).c32.im;
   chi2.c3.im=
      (*u2dn).c31.im*(*u3dn).c12.re+(*u2dn).c31.re*(*u3dn).c12.im+ 
      (*u2dn).c32.im*(*u3dn).c22.re+(*u2dn).c32.re*(*u3dn).c22.im+
      (*u2dn).c33.im*(*u3dn).c32.re+(*u2dn).c33.re*(*u3dn).c32.im;

   _su3_inverse_multiply(psi1,(*u1dn),chi1);
   _su3_inverse_multiply(psi2,(*u1dn),chi2);      
}
   

void staples(int ix,int mu,su3 *v)
{
   int i,nu,ixpmu,ixpnu,ixmnu,ixpmumnu;

   *v=v0;
   ixpmu=iup[ix][mu];

   for (i=1;i<4;i++)
   {
      nu=(mu+i)&0x3;
      ixpnu=iup[ix][nu];
      ixmnu=idn[ix][nu];
      ixpmumnu=idn[ixpmu][nu];
      
      u1up=pu[ix][nu];
      u2up=pu[ixpnu][mu];
      u3up=pu[ixpmu][nu];   
   
      u1dn=pu[ixmnu][nu];
      u2dn=pu[ixmnu][mu];
      u3dn=pu[ixpmumnu][nu];   

      up_staple();
      add_to_v(v);      
      dn_staple();
      add_to_v(v);
   }
}

#endif
